#include <math.h>
#include <stdio.h>
#ifdef __GNUC__
#include <curses.h>
#define MAX_LINES 250
#else
#include <direct.h>
#include <dos.h>
#include <conio.h>
#define COLS 80
#define LINES 25
#define MAX_LINES 25
#endif
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include "watdefs.h"
#include "afuncs.h"
#include "comets.h"
#include "mpc_obs.h"
#include "date.h"

extern double planet_mass[];
extern int perturbers;
extern double stepsize;

#define GAUSS_K .01720209895
#define SOLAR_GM (GAUSS_K * GAUSS_K)
#define PI 3.14159265358979323846

#define EARTH_MAJOR_AXIS 6378140.
#define EARTH_MINOR_AXIS 6356755.
#define EARTH_AXIS_RATIO (EARTH_MINOR_AXIS / EARTH_MAJOR_AXIS)

int find_best_fit_planet( const double jd, const double *ivect,
                                 double *rel_vect, const int included);
double calc_absolute_magnitude( const OBSERVE FAR *obs, int n_obs);
void integrate_orbit( double *orbit, double t0, double t1);
double compute_rms( const OBSERVE FAR *obs, int n_obs, int method);
int herget_method( OBSERVE FAR *obs, int n_obs, double r1, double r2,
                              double *orbit, double *d_r1, double *d_r2);
int adjust_herget_results( OBSERVE FAR *obs, int n_obs, double *orbit);
void improve_parabolic( OBSERVE FAR *obs, int n_obs, double *orbit, double epoch);
void full_improvement( OBSERVE FAR *obs, int n_obs, double *orbit,
                  double epoch, const char *limited_orbit);
int set_locs( const double *orbit, double t0, OBSERVE FAR *obs, int n_obs);
void observation_summary_data( char *obuff, const OBSERVE FAR *obs,
                                const int n_obs);        /* orb_func.cpp */
double DLL_FUNC decimal_day_to_dmy( double jd, long *year, int *month);
int DLL_FUNC elements_in_mpc_format( char *obuff, const ELEMENTS *elem,
                const char *obj_id, const int is_cometary, const int format);
int get_r1_and_r2( const int n_obs, const OBSERVE FAR *obs,
                             double *r1, double *r2);    /* orb_func.cpp */
int initial_orbit( OBSERVE FAR *obs, int n_obs, double *orbit);
int ephemeris_in_a_file( const char *filename, const double *orbit,
         const double epoch_jd, const double jd_start, const double step,
         const double ht_in_meters, const double abs_mag,
         const double lat, const double lon, const int n_steps,
         const int is_comet);             /* ephem0.cpp */
int write_residuals_to_file( const char *filename, const int n_obs,
        const OBSERVE FAR *obs_data, const int short_form);   /* ephem0.cpp */
static void show_a_file( const char *filename);

#ifndef __GNUC__
int getnstr( char *buff, const int max_len)
{
   gets( buff);
   return( 0);
}
#endif

/* Here's a simplified example of the use of the 'ephemeris_in_a_file'
   function... nothing fancy,  but it shows how it's used.  To spare you
   entering lat/lon/alt each time,  I've hardwired these: */

static void create_ephemeris( const double *orbit, const double epoch_jd,
         const double abs_mag, const int is_comet)
{
   double day, jd_start, step;
   double lon = -69.9 * PI / 180.;   /* Lat/lon of Bowdoinham,  Maine, US */
   double lat =  44.01 * PI / 180.;
   double ht_in_meters = 100.;         /* meters */
   int month, year, n_steps;
   const char *ephem_filename = "ephemeri.dat";
   char buff[100];

   printf( "Enter start of ephemeris in D M Y (ex: 25 12 2001): " );
   getnstr( buff, sizeof( buff));
   sscanf( buff, "%lf %d %d", &day, &month, &year);
   jd_start = (double)dmy_to_day( 0, month, (long)year, 0) + day - .5;
   printf( "%lf %d %d = JD %.5lf\n\r", day, month, year, jd_start);
   printf( "Enter step size in days: ");
   getnstr( buff, sizeof( buff));
   sscanf( buff, "%lf", &step);
   printf( "Enter number of steps: ");
   getnstr( buff, sizeof( buff));
   sscanf( buff, "%d", &n_steps);

   if( ephemeris_in_a_file( ephem_filename, orbit,
         epoch_jd, jd_start, step, ht_in_meters, abs_mag,
         lat, lon, n_steps, is_comet))
      {
      printf( "Ephemeris generation failed!\n\r");
      getch( );
      }
   else
      show_a_file( ephem_filename);
}

/* select_object_in_file( ) uses the find_objects_in_file( ) function to
   get a list of all the objects listed in a file of observations.  It
   prints out their IDs and asks the user to hit a key to select one.
   The name of that object is then copied into the 'obj_name' buffer.

   At present,  it's used only in main( ) when you first start DOS_FIND,
   so you can select the object for which you want an orbit.         */

int select_object_in_file( const char *filename, char *obj_name)
{
   int n_ids;
   char *ids = find_objects_in_file( filename, &n_ids, 0);

   *obj_name = '\0';
   if( ids && n_ids)
      {
      char choice = -1;
      int i;

      while( choice < 0 || choice >= n_ids)
         {
         for( i = 0; i < n_ids; i++)
            {
            char buff[20], *id_data = ids + i * 16;

            get_object_name( buff, id_data);
            printf( "%c: %s\n\r", 'A' + i, buff);
            }
         if( n_ids == 1)
            choice = 0;       /* _only_ choice in this file */
         else
            {
            printf( "Hit the letter for your choice: ");
            choice = getch( );
            if( choice >= 'A' && choice <= 'Z')
               choice += 'a' - 'A';
            if( choice >= 'a' && choice < 'a' + n_ids)
               choice -= 'a';
            else
               choice = -1;
            if( choice < 0 || choice >= n_ids)
               printf( "Not a valid choice.  Try again:\n\r");
            }
         }
      get_object_name( obj_name, ids + choice * 16);
      free( ids);
      }
   return( n_ids);
}

/* The set_up_screen( ) function takes just about all information about the
   current object,  observations,  and the state of the orbit as determined
   so far,  and builds up a "screen" buffer full of data about it.  That
   buffer can then just be dumped to the console.  See main( ) for details
   as to its use. */

int set_up_screen( const int n_obs, const OBSERVE FAR *obs, const int curr_obs,
               const double r1, const double r2,
               const int is_cometary, char *screen)
{
   int i, line_start;
   int n_obs_shown = LINES - 19;

   memset( screen, 0, 7 * 80);                 /* clear top seven... */
   memset( screen + 17 * 80, 0, (LINES - 17) * 80);  /* ...and last lines */
   sprintf( screen,
 "R1:%9.5lf       [ Herget step ]   [ Make ephemeris ]   [ Save Residuals ]",
         r1);
   sprintf( screen + 80,
 "R2:%9.5lf       [ Full step   ]   [ save eLements  ]   [   exit (Esc)   ]",
         r2);
   sprintf( screen + 160,
            "                    (%c) Cometary      step siZe:%6.2lf days",
                  (is_cometary ? 'o' : ' '), stepsize);
   strcpy( screen + 240, "Perturbers:");
   screen += 320;

   strcpy( screen,
                "  ( ) 1 Mer   ( ) 3 Ear   ( ) 5 Jup   ( ) 7 Ura   ( ) 9 Plu");
   strcpy( screen + 80,
                "  ( ) 2 Ven   ( ) 4 Mar   ( ) 6 Sat   ( ) 8 Nep   ( ) 0 Lun");
   for( i = 0; i < 10; i++)
      if( (perturbers >> (i + 1)) & 1)
         {
         char *tptr = screen + (i & 1) * 80 + 3;   /* get on right line... */

         tptr[12 * (i / 2)] = 'o';        /* Mark it,  like a radio button */
         }
   screen += 240;       /* skip three lines */
   strcpy( screen, "Orbital elements:");
   screen += 80;

   screen += 9 * 80;    /* nine lines used for orbital elements */
   screen += 80;        /* skip a line,  then: */

   if( n_obs_shown < 0)
      n_obs_shown = 0;
   line_start = curr_obs - n_obs_shown / 2 - 1;
   if( line_start > n_obs - n_obs_shown)
      line_start = n_obs - n_obs_shown;
   if( line_start < 0)
      line_start = 0;

   for( i = 0; i < n_obs_shown; i++, screen += 80)
      if( line_start + i < n_obs)
         {
         sprintf( screen + 1, "%2d: ", (line_start + i + 1) % 100);
         format_observation( obs + line_start + i, screen + 5, 0);
         screen[81] = '\0';
         if( line_start + i != curr_obs)
            screen[0] = ' ';
         else
            {
            screen[0] = '[';
            strcat( screen, "]");
            }
         }
   put_observer_data_in_text( obs[curr_obs].mpc_code, screen);
   return( 0);
}

static double get_epoch_of_last_included_obs( OBSERVE FAR *obs, int n_obs)
{
   int i;

   for( i = 0; i < n_obs - 1 && !obs[i].is_included; i++)
      ;
   return( obs[i].jd);
}

static void show_a_file( const char *filename)
{
   FILE *ifile = fopen( filename, "rb");
   char buff[100];
   int line_no = 0, c = 0;

   while( fgets( buff, 80, ifile) && c != 27)
      {
      printf( "%s\r", buff);
      line_no++;
      if( !(line_no % (LINES - 1)))
         {
         printf( "Hit a key:");
         c = getch( );
         printf( "\n\r");
         }
      }
   fclose( ifile);
   if( c != 27 && (line_no % (LINES - 1)))
      {
      printf( "Hit a key:");
      c = getch( );
      printf( "\n\r");
      }
}

/* main( ) begins by using the select_object_in_file( ) function (see above)
   to determine which object is going to be analyzed in the current 'run'.
   It loads up the observations for the object to be analyzed,
   and computes an initial guess for the orbit.

   It then goes into a loop,  setting up and displaying a screenful of
   data,  then asking the user to hit a key.  Based on which key is hit,
   an action such as taking an Herget step or a "full step",  or resetting
   the epoch,  or quitting,  is taken. */

void main( int argc, char **argv)
{
   char obj_name[30], tbuff[90], orbit_constraints[90];
   int n_ids, c = 1, precision = 0;
   OBSERVE FAR *obs;
   int n_obs, i, is_cometary = 0, curr_obs;
   double epoch_shown, curr_epoch, orbit[6];
   double r1 = 1., r2 = 1.;
   char *buff = (char *)malloc( 80 * MAX_LINES);
   int update_element_display = 1;

#ifdef __GNUC__
   initscr( );
   cbreak( );
// noecho( );
// clear( );
#endif
   setvbuf( stdout, NULL, _IONBF, 0);
   *orbit_constraints = '\0';
   n_ids = select_object_in_file( argv[1], obj_name);
   if( !*obj_name)         /* no objects found,  or none chosen */
      exit( -1);
   obs = load_observations( argv[1], obj_name, &n_obs, 0);
   for( i = 0; i < n_obs - 1 && !obs[i].is_included; i++)
      ;
   curr_obs = i;
   curr_epoch = get_epoch_of_last_included_obs( obs, n_obs);
   epoch_shown = floor( curr_epoch / 20.) * 20. + .5;
   curr_epoch = epoch_shown;

   if( initial_orbit( obs, n_obs, orbit))    /* fallback position: */
      herget_method( obs, n_obs, 1., 1., orbit, NULL, NULL);
   else
      get_r1_and_r2( n_obs, obs, &r1, &r2);    /* orb_func.cpp */
   while( c != 27)
      {
      set_up_screen( n_obs, obs, curr_obs,
                   r1, r2, is_cometary, buff);
      if( update_element_display)
         {
         double rel_orbit[6], orbit2[6];
         int planet_orbiting, n_lines;
         ELEMENTS elem;
         char *tptr, *tbuff = (char *)malloc( 80 * 9);

         memcpy( orbit2, orbit, 6 * sizeof( double));
         integrate_orbit( orbit2, curr_epoch, epoch_shown);
         planet_orbiting = find_best_fit_planet( epoch_shown, orbit2, rel_orbit, -1);
         calc_classical_elements( &elem, rel_orbit, epoch_shown, 1,
                                    SOLAR_GM * planet_mass[planet_orbiting]);
         elem.central_obj = planet_orbiting;
         elem.epoch = epoch_shown;
         elem.abs_mag = calc_absolute_magnitude( obs, n_obs);
         elem.slope_param = .15;
         n_lines = elements_in_mpc_format( tbuff, &elem, obs->obj_id,
                                              is_cometary, precision);
         tptr = tbuff;
         memset( buff + 8 * 80, 0, 10 * 80);
         for( i = 0; i < n_lines; i++)
            {
            strcpy( buff + (i + 8) * 80, tptr);
            if( *tptr)
               tptr += strlen( tptr) + 1;
            }
         observation_summary_data( buff + (n_lines + 8) * 80, obs, n_obs);
         free( tbuff);
         update_element_display = 0;
         }

      printf( "\n\r");
      for( i = 0; i < LINES; i++)
         {
         memcpy( tbuff, buff + i * 80, 80);
         tbuff[80] = '\0';
#ifdef __GNUC__
         if( COLS < 80)
            tbuff[COLS] = '\0';
#endif
         printf( "%s", tbuff);
         if( i != LINES - 1 && strlen( tbuff) < COLS)
            printf( "\n\r");
         }
      c = getch( );
      printf( "\n\r");
      if( c >= '1' && c <= '9')
         perturbers ^= (1 << (c - '0'));
      else switch( c)
         {
         case '0':
            perturbers ^= 1024;
            break;
         case 'c': case 'C':
            is_cometary ^= 1;
            break;
         case 'q': case 'Q':
            if( curr_obs > 0)
               curr_obs--;
            break;
         case 'a': case 'A':
            if( curr_obs < n_obs - 1)
               curr_obs++;
            break;
         case 'x': case 'X':
            obs[curr_obs].is_included ^= 1;
            break;
         case 'd': case 'D':
            {
            static const char *resid_filename = "residual.dat";

            write_residuals_to_file( resid_filename,  n_obs, obs, (c == 'D'));
            show_a_file( resid_filename);
            }
            break;
         case 'e': case'E':
            printf( "Enter new epoch (as a JD):");
            getnstr( tbuff, sizeof( tbuff));
            sscanf( tbuff, "%lf", &epoch_shown);
            update_element_display = 1;
            break;
         case 'f': case 'F':
            {
//          double new_epoch = get_epoch_of_last_included_obs( obs, n_obs);
            double new_epoch = epoch_shown;

            if( curr_epoch != new_epoch)
               {
               integrate_orbit( orbit, curr_epoch, new_epoch);
               curr_epoch = new_epoch;
               }
            if( is_cometary)
               improve_parabolic( obs, n_obs, orbit, curr_epoch);
            else
               full_improvement( obs, n_obs, orbit, curr_epoch,
                                          orbit_constraints);
            get_r1_and_r2( n_obs, obs, &r1, &r2);
            update_element_display = 1;
            }
            break;
         case 'h': case 'H':
            {
            double d_r1, d_r2;

            is_cometary = 0;
            herget_method( obs, n_obs, r1, r2, orbit, &d_r1, &d_r2);
            r1 += d_r1;
            r2 += d_r2;
            herget_method( obs, n_obs, r1, r2, orbit, NULL, NULL);
            curr_epoch = get_epoch_of_last_included_obs( obs, n_obs);
            update_element_display = 1;
            }
            break;
         case 'l': case 'L':
            printf( "Enter limits on the orbit (e.g.,  'e=0' or 'q=2.3'):\n");
            getnstr( orbit_constraints, sizeof( orbit_constraints));
            break;
         case 'o': case 'O':
            {
            char filename[80];
            FILE *ofile;

            printf( "Enter filename for saving elements: ");
            getnstr( tbuff, sizeof( tbuff));
            sscanf( tbuff, "%s", filename);
            ofile = fopen( filename, "wb");
            for( i = 7; i < 17; i++)
               fprintf( ofile, "%s\n", buff + i * 80);
            fclose( ofile);
            }
            break;
         case 'p': case 'P':
            precision ^= 2;
            update_element_display = 1;
            break;
         case 'm': case 'M':
            create_ephemeris( orbit, curr_epoch,
                   calc_absolute_magnitude( obs, n_obs),
                   obs->is_comet);
            break;
         case 'r': case 'R':
            printf( "Enter new R1, R2: ");
            getnstr( tbuff, sizeof( tbuff));
            sscanf( tbuff, "%lf %lf", &r1, &r2);
            break;
         case 'v': case 'V':
            {
            double vaisala_dist;

            printf( "Enter peri/apohelion distance: ");
            getnstr( tbuff, sizeof( tbuff));
            sscanf( tbuff, "%lf", &vaisala_dist);
            herget_method( obs, n_obs, -vaisala_dist, 0., orbit, NULL, NULL);
            adjust_herget_results( obs, n_obs, orbit);
            curr_epoch = get_epoch_of_last_included_obs( obs, n_obs);
            get_r1_and_r2( n_obs, obs, &r1, &r2);
            update_element_display = 1;
            }
            break;
         case 'z': case 'Z':
            printf( "Enter new step size (in days): ");
            getnstr( tbuff, sizeof( tbuff));
            sscanf( tbuff, "%lf", &stepsize);
            break;
         case 27:
            break;
         default:
            show_a_file( "dos_help.doc");
            break;
         }
      }
#ifdef __GNUC__
   endwin( );
#endif
}
