#include <math.h>
#include <stdio.h>
#include <string.h>
#include "watdefs.h"
#include "afuncs.h"
#include "lunar.h"
#include "date.h"
#include "comets.h"
#include "mpc_obs.h"

#define J2000 2451545.0
#define EARTH_MAJOR_AXIS 6378140.
#define EARTH_MINOR_AXIS 6356755.
#define EARTH_AXIS_RATIO (EARTH_MINOR_AXIS / EARTH_MAJOR_AXIS)
#define AU_IN_KM 1.4959787e+8
#define AU_PER_DAY (86400. * 299792.456 / AU_IN_KM)
#define EARTH_MAJOR_AXIS_IN_AU (EARTH_MAJOR_AXIS / (1000. * AU_IN_KM))
#define PI 3.141592653589793238462643383279502884197169399375105

double DLL_FUNC decimal_day_to_dmy( double jd, long *year, int *month); /* orb_func.c */
void integrate_orbit( double *orbit, double t0, double t1);
double calc_obs_magnitude( int is_comet, double obj_sun, double obj_earth,
                                double earth_sun);  /* orb_func.cpp */
int ephemeris_in_a_file( const char *filename, const double *orbit,
         const double epoch_jd, const double jd_start, const double step,
         const double ht_in_meters, const double abs_mag,
         const double lat, const double lon, const int n_steps,
         const int is_comet);             /* ephem0.cpp */
int write_residuals_to_file( const char *filename, const int n_obs,
        const OBSERVE FAR *obs_data, const int short_form);   /* ephem0.cpp */
void put_observer_data_in_text( const char FAR *mpc_code, char *buff);
int earth_lunar_posn( const double jd, double FAR *earth_loc, double FAR *lunar_loc);

int ephemeris_in_a_file( const char *filename, const double *orbit,
         const double epoch_jd, const double jd_start, const double step,
         const double ht_in_meters, const double abs_mag,
         const double lat, const double lon, const int n_steps,
         const int is_comet)
{
   double orbi[6], curr_jd, rho_cos_lat, rho_sin_lat;
   double prev_ephem_t = epoch_jd;
   const double u = atan( sin( lat) * EARTH_AXIS_RATIO / cos( lat));
   int i, j;
   FILE *ofile = fopen( filename, "wb");

   if( !ofile)
      return( -1);
   rho_sin_lat = EARTH_AXIS_RATIO * sin( u) +
                           (ht_in_meters / EARTH_MAJOR_AXIS) * sin( lat);
   rho_cos_lat = cos( u) + (ht_in_meters / EARTH_MAJOR_AXIS) * cos( lat);
   rho_sin_lat *= EARTH_MAJOR_AXIS_IN_AU;
   rho_cos_lat *= EARTH_MAJOR_AXIS_IN_AU;

   memcpy( orbi, orbit, 6 * sizeof( double));
   curr_jd = jd_start;
   for( i = 0; i < n_steps; i++)
      {
      double geocent[3], ephemeris_t, planet_loc[3], hr_ang, posn_ang;
      double topo[6], r = 0., solar_r = 0., obs_posn[3];
      const double sin_obliq_2000 = .397777156;
      const double cos_obliq_2000 = .917482062;
      double day, temp, ra, dec, sec, earth_r = 0.;
      int month, hr, min, deg, dec_sign = '+';
      long year;
      char buff[80], ra_buff[80], dec_buff[80];

      ephemeris_t = curr_jd + td_minus_ut( curr_jd) / 86400.;
      integrate_orbit( orbi, prev_ephem_t, ephemeris_t);
      prev_ephem_t = ephemeris_t;
      earth_lunar_posn( ephemeris_t, planet_loc, NULL);
      FMEMCPY( obs_posn, planet_loc, 3 * sizeof( double));
      for( j = 0; j < 3; j++)
         {
         topo[j] = orbi[j] - obs_posn[j];
         r += topo[j] * topo[j];
         earth_r += obs_posn[j] * obs_posn[j];
         }
      earth_r = sqrt( earth_r);
      memcpy( topo + 3, topo, 3 * sizeof( double));
      r = sqrt( r);
                              /* use this distance to do light-time lag: */
      for( j = 0; j < 3; j++)
         topo[j] -= orbi[j + 3] * r / AU_PER_DAY;
      r = 0.;

                  /* rotate topo from ecliptic to equatorial */
      temp    = topo[2] * cos_obliq_2000 + topo[1] * sin_obliq_2000;
      topo[1] = topo[1] * cos_obliq_2000 - topo[2] * sin_obliq_2000;
      topo[2] = temp;

      temp    = topo[5] * cos_obliq_2000 + topo[4] * sin_obliq_2000;
      topo[4] = topo[4] * cos_obliq_2000 - topo[5] * sin_obliq_2000;
      topo[5] = temp;

      {
      double ra_dec[4];

      for( j = 0; j < 2; j++)
         {
         double *tptr = topo + j * 3;

         temp = tptr[0] * tptr[0] + tptr[1] * tptr[1] + tptr[2] * tptr[2];
         ra_dec[j + j] = atan2( tptr[1], tptr[0]);
         ra_dec[j + j + 1] = asin( tptr[2] / sqrt( temp));
         }
      calc_dist_and_posn_ang( ra_dec, ra_dec + 2, &temp, &posn_ang);
      posn_ang = PI + PI - posn_ang;      /* weird reversal of direction */
      }

      hr_ang = green_sidereal_time( curr_jd) + lon;
      geocent[0] = rho_cos_lat * cos( hr_ang);
      geocent[1] = rho_cos_lat * sin( hr_ang);
      geocent[2] = rho_sin_lat;
      for( j = 0; j < 3; j++)
         {
         topo[j] -= geocent[j];
         r += topo[j] * topo[j];
         solar_r += orbi[j] * orbi[j];
         }
      r = sqrt( r);
      solar_r = sqrt( solar_r);

      ra = atan2( topo[1], topo[0]) * 12. / PI;
      if( ra < 0.) ra += 24.;
      hr = (int)ra;
      min = (int)((ra - (double)hr) * 60.);
      sec =       (ra - (double)hr) * 3600. - 60. * (double)min;
      sprintf( ra_buff, "%02d %02d %6.3lf", hr, min, sec);
      if( ra_buff[6] == ' ')        /* leading zero */
         ra_buff[6] = '0';

      dec = asin( topo[2] / r) * 180. / PI;
      if( dec < 0.)
         {
         dec = -dec;
         dec_sign = '-';
         }
      deg = (int)dec;
      min = (int)( (dec - (double)deg) * 60.);
      sec =        (dec - (double)deg) * 3600. - (double)min * 60.;
      sprintf( dec_buff, "%c%02d %02d %5.2lf", dec_sign, deg, min, sec);
      if( dec_buff[7] == ' ')        /* leading zero */
         dec_buff[7] = '0';

      day = decimal_day_to_dmy( curr_jd, &year, &month);
      sprintf( buff, "%8.5lf %02d %02ld  %s   %s  %.4lf %.4lf %.1lf",
            day, month, year % 100L, ra_buff, dec_buff, r, solar_r,
            posn_ang * 180. / PI);
      if( step == floor( step + .5))
         {        /* integer step size:  may clip trailing zeroes */
         for( j = 7; buff[j] == '0'; j--)
            ;
         if( buff[j] == '.')
            j--;
         memmove( buff + j + 1, buff + 8, strlen( buff + 7));
         }
      if( abs_mag)
         {
         double curr_mag = abs_mag + calc_obs_magnitude( is_comet,
                              solar_r, r, earth_r);      /* orb_func.cpp */

         sprintf( buff + strlen( buff), " %.1lf", curr_mag + .05);
         }

      fprintf( ofile, "%s\n", buff);

      curr_jd += step;
      }
   fclose( ofile);
   return( 0);
}

/* format_observation( ) takes an observation and produces text for it,
   suitable for display on a console (DOS_FIND) or in a Windoze scroll
   box (FIND_ORB).   */

void format_observation( const OBSERVE FAR *obs, char *text, const int use_tabs)
{
   const double pi = 3.141592653589793238462643383279502884197;
   double ang = obs->ra * 12. / pi;
   double residuals[2];
   int deg, hr, min, i;
   int month;
   long year;
   double day, univ_t;
   char tbuff[10];
   static char *first_format[2] = {
                  "%2ld %2d %8.5lf %c %s %02d %02d %5.2lf  ",
                  "%2ld\t%2d\t%8.5lf\t%c\t%s\t%02d\t%02d\t%5.2lf\t" };
   static char *second_format[2] = {
                  "%02d %02d %4.1lf %6.2lf %6.2lf %7.3lf%7.3lf",
                  "%02d\t%02d\t%4.1lf\t%6.2lf\t%6.2lf\t%6.3lf\t%6.3lf" };

   if( ang < 0.) ang += 24.;
      hr = (int)ang;
   min = (int)((ang - hr) * 60.);
   ang -= hr + (double)min / 60.;
   univ_t = obs->jd - td_minus_ut( obs->jd) / 86400.;
   day = decimal_day_to_dmy( univ_t, &year, &month);

   FSTRCPY( tbuff, obs->mpc_code);
   if( use_tabs != 2)
      {
      char format[60];

      strcpy( format, first_format[use_tabs]);
                             /* high-precision time modifications: */
      format[10] = (char)( '8' + ( obs->precision & 1));
      format[12] = (char)( '5' + ( obs->precision & 1));
                             /* high-precision RA modifications: */
      format[33] = (char)( '5' + ( obs->precision & 2) / 2);
      format[35] = (char)( '2' + ( obs->precision & 2) / 2);
      sprintf( text, format,
                   year % 100L, month, day,
                   (obs->is_included ? ' ' : 'X'), tbuff,
                   hr, min, ang * 3600.);
      }
   else        /* 'short' MPC format: */
      sprintf( text, "%02ld%02d%02d %s", year % 100L, month, (int)day, tbuff);
   text += strlen( text);

   ang = obs->dec * 180. / pi;
   if( use_tabs != 2)
      if( ang < 0.)
         {
         ang = -ang;
         *text++ = '-';
         }
      else
         *text++ = '+';
   deg = (int)ang;
   min = (int)((ang - deg) * 60.);
   ang -= deg + (double)min / 60.;
   residuals[0] = 3600. * (180./pi) * (obs->ra - obs->computed_ra);
   residuals[1] = 3600. * (180./pi) * (obs->dec - obs->computed_dec);
   if( use_tabs != 2)
      {
      char format[60];

      strcpy( format, second_format[use_tabs]);

                             /* high-precision dec modifications: */
      format[11] = (char)( '4' + ( obs->precision & 4) / 4);
      format[13] = (char)( '1' + ( obs->precision & 4) / 4);
      sprintf( text, format,
                          deg, min, ang * 3600., residuals[0], residuals[1],
                          obs->r, obs->solar_r);
      if( residuals[0] < .001 && residuals[0] > -.001
                  && residuals[1] < .001 && residuals[1] > -.001)
         {
         char *text_loc = text + 11;

         if( obs->precision & 4)
            text_loc++;
         memcpy( text_loc, "  ----", 6);
         memcpy( text_loc + 7, "  ----", 6);
         }
      }
   else        /* 'short' MPC format */
      {
      for( i = 0; i < 2; i++)
         {
         double zval = fabs( residuals[i]);

         if( zval > 99.9)
            strcpy( text, "-----");
         else if( zval > .999)
            sprintf( text, "%5.1lf", zval);
         else
            {
            sprintf( text, "%5.2lf", zval);
            text[1] = ' ';
            }
         if( !strcmp( text, "  .00"))
            text[5] = ' ';
         else
            text[5] = (residuals[i] > 0. ? '+' : '-');
         text += 6;
         }
      text[-12] = (obs->is_included ? ' ' : '(');
      *text++ = (obs->is_included ? ' ' : ')');
      *text++ = '\0';
      }
}

int write_observer_data_to_file( FILE *ofile, const int n_obs,
                             const OBSERVE FAR *obs_data)
{
   int stations[100], n_stations = 0, i, j;

   for( i = 0; i < n_obs; i++)
      {
      int match_found = 0;

      for( j = 0; j < n_stations && !match_found; j++)
         match_found = !FSTRCMP( obs_data[i].mpc_code,
                                obs_data[stations[j]].mpc_code);
      if( !match_found)    /* new one:  add it to the array */
         stations[n_stations++] = i;
      }
               /* now do a simple bubblesort: */
   for( i = 0; i < n_stations; i++)
      for( j = 0; j < i; j++)
         if( FSTRCMP( obs_data[stations[i]].mpc_code,
                     obs_data[stations[j]].mpc_code) < 0)
            {
            int temp = stations[i];

            stations[i] = stations[j];
            stations[j] = temp;
            }

   for( i = 0; i < n_stations; i++)
      {
      char buff[100], tbuff[10];

      FSTRCPY( tbuff, obs_data[stations[i]].mpc_code);
      put_observer_data_in_text( tbuff, buff);
      fprintf( ofile, "(%s) %s\n", tbuff, buff);
      }
   return( 0);
}


int write_residuals_to_file( const char *filename, const int n_obs,
                             const OBSERVE FAR *obs_data, const int short_form)
{
   FILE *ofile = fopen( filename, "w");
   int rval = 0;

   if( ofile )
      {
      char buff[100];
      int number_lines = (n_obs + 2) / 3;
      int i;

      if( short_form)
         for( i = 0; i < number_lines * 3; i++)
            {
            int num = (i % 3) * number_lines + i / 3;
            OBSERVE FAR *obs = ((OBSERVE FAR *)obs_data) + num;

            if( num < n_obs)
               format_observation( obs, buff, 2);
            else
               *buff = '\0';
            fprintf( ofile, "%s%s", buff, (i % 3 == 2) ? "\n" : "   ");
            }
      else
         for( i = 0; i < n_obs; i++)
            {
            format_observation( obs_data + i, buff, 0);
            fprintf( ofile, "%s\n", buff);
            }
      fprintf( ofile, "\nStation data:\n");
      write_observer_data_to_file( ofile, n_obs, obs_data);
      fclose( ofile);
      }
   else                    /* file not opened */
      rval = -1;
   return( rval);
}

