#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#ifdef _WINDOWS
#include <windows.h>
#endif
#include "watdefs.h"
#include "comets.h"
#include "mpc_obs.h"
#include "lunar.h"
#include "afuncs.h"
#include "date.h"
#include "jpleph.h"


#define J2000 2451545.0
#define PI 3.141592653589793238462643383279502884197
#define EARTH_MAJOR_AXIS 6378140.
#define EARTH_MINOR_AXIS 6356755.
#define EARTH_AXIS_RATIO (EARTH_MINOR_AXIS / EARTH_MAJOR_AXIS)
#define AU_IN_KM 1.4959787e+8
#define EARTH_MAJOR_AXIS_IN_AU (EARTH_MAJOR_AXIS / (1000. * AU_IN_KM))
#define EARTH_MINOR_AXIS_IN_AU (EARTH_MINOR_AXIS / (1000. * AU_IN_KM))
#define EARTH_MOON_BARYCENTER_FACTOR (82.300679)

double DLL_FUNC decimal_day_to_dmy( double jd, long *year, int *month);
void set_obs_vect( OBSERVE FAR *obs);        /* mpc_obs.h */
int planet_posn( const int planet_no, const double jd, double *vect_2000);

/* MPC reports store RA and dec in base 60,  Babylonian style. */

static double read_base_60( char *ibuff)
{
   double rval = atof( ibuff) + atof( ibuff + 3) / 60. +
                                       atof( ibuff + 6) / 3600.;

   return( rval);
}

/* The following function paws through the STATIONS.TXT file,  looking for the
   observer code in question.  When it finds it,  it just copies the entire
   line into buff,  performing no analysis... that's the caller's job. */

int get_observer_data( const char FAR *mpc_code, char *buff)
{
   FILE *ifile;

   ifile = fopen( "stations.txt", "rb");
   if( !ifile)
      return( -1);
   while( fgets( buff, 100, ifile))
      if( !FMEMCMP( buff, mpc_code, 3) && buff[3] == ' ')
         {
         fclose( ifile);
         return( 0);
         }
   fclose( ifile);
   return( -2);
}


/* In an MPC astrometric report line,  the name can be stored in assorted
   highly scrambled and non-intuitive ways.  Those ways _are_ documented
   on the MPC Web site,  which is probably the best place to look to
   understand why the following bizarre code does what it does. */

int get_object_name( char *obuff, const char *ibuff)
{
   int rval = -1;

   if( ibuff[4] == 'S')    /* Possible natural satellite... let's check: */
      if( strchr( "MVEJSUNP", *ibuff) && isdigit( ibuff[1])
                  && isdigit( ibuff[2]) && isdigit( ibuff[3]) &&
                  !memcmp( ibuff + 5, "       ", 7))
         {
         const char *planet_names[8] = { "Venus", "Earth", "Mars", "Jupiter",
                  "Saturn", "Uranus", "Neptune", "Pluto" };
         const char *roman_numerals[10] = { "", "I", "II", "III", "IV",
                     "V", "VI", "VII", "VIII", "IX" };
         int i;

         for( i = 0; i < 8; i++)
            if( planet_names[i][0] == *ibuff)
               {
               strcpy( obuff, planet_names[i]);
               strcat( obuff, " ");
               }
         for( i = atoi( ibuff + 1); i >= 10; i -= 10)
            strcat( obuff, "X");
         strcat( obuff, roman_numerals[i]);
         rval = 0;
         }
      else if( strchr( "MVEJSUNP", ibuff[8]) && isdigit( ibuff[6])
               && isdigit( ibuff[7]) && isdigit( ibuff[9])
               && isdigit( ibuff[10]) && ibuff[11] == '0'
               && !memcmp( ibuff + 1, "   ", 3) && ibuff[5] >= 'J'
               && ibuff[5] <= 'Z')
         {
         sprintf( obuff, "S/%d%c%c", 20 + ibuff[5] - 'K', ibuff[6], ibuff[7]);
         obuff[6] = ' ';
         obuff[7] = ibuff[8];       /* planet identifier */
         obuff[8] = ' ';
         if( ibuff[9] > '0')        /* double-digit ID (unlikely,  but...) */
            {
            obuff[9] = ibuff[9];
            obuff[10] = ibuff[10];
            obuff[11] = '\0';
            }
         else                 /* more usual single-digit case: */
            {
            obuff[9] = ibuff[10];
            obuff[10] = '\0';
            }
         rval = 0;
         }

   if( ibuff[4] >= '0' && ibuff[4] <= '9')         /* numbered asteroid */
      if( isdigit( ibuff[0]) && isdigit( ibuff[1]) &&
          isdigit( ibuff[2]) && isdigit( ibuff[3]) &&
          !memcmp( ibuff + 5, "       ", 7))
         {
         *obuff++ = '(';
         while( *ibuff == '0')         /* skip leading zeroes */
            ibuff++;
         while( *ibuff >= '0' && *ibuff <= '9')     /* read digits... */
            *obuff++ = *ibuff++;
         *obuff++ = ')';
         *obuff++ = '\0';
         rval = 0;
         }

   if( strchr( "PCDXA", ibuff[4]))
      {
      *obuff++ = ibuff[4];
      *obuff++ = '/';
      if( ibuff[5] == ' ')             /* numbered object */
         {
         while( *ibuff == '0')         /* skip leading zeroes */
            ibuff++;
         while( *ibuff >= '0' && *ibuff <= '9')   /* read digits... */
            *obuff++ = *ibuff++;
         *obuff++ = '\0';
         rval = 0;
         }
      }
   if( rval)
      if( ibuff[5] == 'J' || ibuff[5] == 'K')
         {
         *obuff++ = ((ibuff[5] == 'J') ? '1' : '2');
         *obuff++ = ((ibuff[5] == 'J') ? '9' : '0');
         *obuff++ = ibuff[6];             /* decade */
         *obuff++ = ibuff[7];             /* year */
         *obuff++ = ibuff[8];
         if( ibuff[11] != '0')
            *obuff++ = ibuff[11];
         if( ibuff[9] != '0')
            if( ibuff[9] >= 'A')
               {
               *obuff++ = (char)( (ibuff[9] - 'A') / 10 + '1');
               *obuff++ = (char)( (ibuff[9] - 'A') % 10 + '0');
               }
            else
               *obuff++ = ibuff[9];
         if( ibuff[10] != '0' || ibuff[9] != '0')
            *obuff++ = ibuff[10];
         *obuff++ = '\0';
         rval = 0;
         }
      else if( ibuff[5] != ' ')  /* Palomar-Leiden,  or temp desig */
         {
         ibuff += 5;
         while( *ibuff != ' ' && *ibuff != '*')
            *obuff++ = *ibuff++;
         *obuff++ = '\0';
         rval = 0;
         }
   return( rval);
}

void equatorial_to_ecliptic( double FAR *vect)
{
   const double sin_obliq_2000 = .397777156;
   const double cos_obliq_2000 = .917482062;
   double temp;

   temp    = vect[2] * cos_obliq_2000 - vect[1] * sin_obliq_2000;
   vect[1] = vect[1] * cos_obliq_2000 + vect[2] * sin_obliq_2000;
   vect[2] = temp;
}

void set_obs_vect( OBSERVE FAR *obs)
{
   obs->vect[0] = cos( obs->ra) * cos( obs->dec);
   obs->vect[1] = sin( obs->ra) * cos( obs->dec);
   obs->vect[2] = sin( obs->dec);
   equatorial_to_ecliptic( obs->vect);
}

/* Given a planet number (Sun=0, Mercury=1, ... Pluto=9) and a JD,
   the following code computes that planet's J2000 equatorial coordinates
   using the PS1996 theory (or,  for the moon,  ELP82). */

const char *get_environment_ptr( const char *env_ptr);     /* mpc_obs.cpp */

int planet_posn( const int planet_no, const double jd, double *vect_2000)
{
   static void *ps_1996_data[9];
   int i, rval = 0;
   static const char *jpl_filename = NULL;
   static void *jpl_eph = NULL;

   if( !planet_no)            /* the sun */
      {
      vect_2000[0] = vect_2000[1] = vect_2000[2] = 0.;
      return( 0);
      }
   if( !jpl_filename)
      {
#ifdef __GNUC__
      jpl_filename = get_environment_ptr( "LINUX_JPL_FILENAME");
#else
      jpl_filename = get_environment_ptr( "JPL_FILENAME");
#endif
      jpl_eph = jpl_init_ephemeris( jpl_filename, NULL, NULL);
      }

   if( jpl_eph)
      {
      double state[6];            /* DE gives both posn & velocity */
      int got_it = 0;

      if( planet_no < 0)          /* flag to unload everything */
         {
         jpl_close_ephemeris( jpl_eph);
         jpl_eph = NULL;
         return( 0);
         }
      else if( planet_no == 10)
         got_it = !jpl_pleph( jpl_eph, jd, 10, 3, state, 0);
      else
         got_it = !jpl_pleph( jpl_eph, jd,
                 (planet_no == 3) ? 13 : planet_no, 11, state, 0);
      if( got_it && planet_no > 0)         /* we're done */
         {
         equatorial_to_ecliptic( state);
         memcpy( vect_2000, state, 3 * sizeof( double));
         return( 0);
         }
      }

   if( planet_no == 10)        /* the moon */
      {
      double tloc[4];

      if( !compute_elp_xyz( NULL, (jd - J2000) / 36525., 0., tloc))
         for( i = 0; i < 3; i++)
            vect_2000[i] = tloc[i] / AU_IN_KM;
      else
         rval = -3;
      return( rval);
      }

   if( planet_no < 0)          /* flag to unload everything */
      {
      for( i = 0; i < 9; i++)
         if( ps_1996_data[i])
            {
            unload_ps1996_series( ps_1996_data[i]);
            ps_1996_data[i] = NULL;
            }
      return( 0);
      }

   if( !ps_1996_data[planet_no])
      ps_1996_data[planet_no] = load_ps1996_series( NULL, jd, planet_no);

   if( !ps_1996_data[planet_no])
      rval = -1;
   else if( get_ps1996_position( jd, ps_1996_data[planet_no], vect_2000, 0))
      {
      unload_ps1996_series( ps_1996_data[planet_no]);
      ps_1996_data[planet_no] = load_ps1996_series( NULL, jd, planet_no);
      if( !ps_1996_data[planet_no])
         rval = -2;
      else if( get_ps1996_position( jd, ps_1996_data[planet_no], vect_2000, 0))
         rval = -3;
      }

   if( !rval)
      equatorial_to_ecliptic( vect_2000);
   else
      {
      FILE *debug_file = fopen( "debug.dat", "ab");

      fprintf( debug_file, "Err %d in planet_posn: %d, JD %lf\n",
                                 rval, planet_no, jd);
      fclose( debug_file);
      }

               /* The 16-bit Windoze version runs low on memory if you */
               /* try to keep everything in memory.  So we free as much */
               /* as is reasonable: */
#ifndef _WIN32
   for( i = 0; i < 9; i++)
      if( ps_1996_data[i] && i != planet_no)
         {
         unload_ps1996_series( ps_1996_data[i]);
         ps_1996_data[i] = NULL;
         }
#endif

   return( rval);
}

int earth_lunar_posn( const double jd, double FAR *earth_loc, double FAR *lunar_loc)
{
   double t_earth[3], t_lunar[3];
   int i;

   planet_posn( 3, jd, t_earth);
   planet_posn( 10, jd, t_lunar);
                     /* earth_loc is the E-M barycenter,  and lunar_loc */
                     /* is geocentric.  Modify earth_loc to be the earth */
                     /* geocenter loc,  and lunar_loc to be heliocentric: */
   for( i = 0; i < 3; i++)
      {
      t_earth[i] -= t_lunar[i] / EARTH_MOON_BARYCENTER_FACTOR;
      t_lunar[i] += t_earth[i];
      if( earth_loc)
         earth_loc[i] = t_earth[i];
      if( lunar_loc)
         lunar_loc[i] = t_lunar[i];
      }
   return( 0);
}

/* parse_observation( ) takes an MPC astrometric observation of the usual
   80-character variety,  and extracts all relevant data from it and puts
   it into the 'obs' structure.  It also computes the ecliptic J2000
   coordinates of the observer at the time of the observation,  using the
   PS1996 and ELP82 theories.                         */

static int parse_observation( OBSERVE FAR *obs, char *buff,
                                              const double weight)
{
   int month = atoi( buff + 20);
   long year = atol( buff + 15);
   double geocent[3];
   static char curr_station[4];
   static double lon, rho_cos_lat, rho_sin_lat;
   double hr_ang, ut;
   char tbuff[100];
   int i, unfound[10][4], n_unfound = 0;

   if( year > 2010L || year < 1600L || month < 1 || month > 12)
      return( -1);

   get_object_name( tbuff, buff);
   FSTRCPY( obs->obj_id, tbuff);
   ut = obs->jd = dmy_to_day( 0, month, year, 0) - .5 + atof( buff + 23);
   obs->jd += td_minus_ut( obs->jd) / 86400.;
   earth_lunar_posn( obs->jd, obs->obs_posn, NULL);

   obs->ra  = read_base_60( buff + 32) * (PI / 12.);
   obs->dec = read_base_60( buff + 45) * (PI / 180.);

   obs->precision = (buff[31] != ' ');
   if( buff[43] != ' ')
      obs->precision |= 2;
   if( buff[55] != ' ')
      obs->precision |= 4;

   obs->is_included = (buff[71] != 'x');
   if( buff[44] == '-')
      obs->dec = -obs->dec;
   if( buff[67] == '.')
      obs->obs_mag = atof( buff + 65);
   else
      obs->obs_mag = 0.;
   obs->is_comet = ( buff[4] != ' ' && (buff[4] < '0' && buff[4] > '9'));
   FMEMCPY( obs->mpc_code, buff + 77, 3);
   obs->mpc_code[3] = '\0';
   if( FMEMCMP( obs->mpc_code, curr_station, 3))
      {
      if( get_observer_data( obs->mpc_code, tbuff))
         {
         for( i = 0; FMEMCMP( unfound[i], obs->mpc_code, 3) && i < n_unfound; i++)
            ;
         if( i == n_unfound && i < 10)
            {
            FMEMCPY( unfound[n_unfound++], obs->mpc_code, 3);
#ifdef _WINDOWS
            sprintf( tbuff, "Didn't find observer %s", obs->mpc_code);
            MessageBox( NULL, tbuff, "FindOrb", MB_OK);
#else
            printf( "WARNING:  didn't find observer %s\n", obs->mpc_code);
#endif
            }
         }
      FMEMCPY( curr_station, obs->mpc_code, 3);
      rho_cos_lat = rho_sin_lat = 0.;
      sscanf( tbuff + 4, "%lf%lf%lf", &lon, &rho_cos_lat, &rho_sin_lat);
//    y /= EARTH_AXIS_RATIO * EARTH_AXIS_RATIO;
      if( lon > 180.)
         lon -= 360.;
      lon *= PI / 180.;
      }
   hr_ang = green_sidereal_time( ut) + lon;
   geocent[0] = EARTH_MAJOR_AXIS_IN_AU * rho_cos_lat * cos( hr_ang);
   geocent[1] = EARTH_MAJOR_AXIS_IN_AU * rho_cos_lat * sin( hr_ang);
   geocent[2] = EARTH_MAJOR_AXIS_IN_AU * rho_sin_lat;
   equatorial_to_ecliptic( geocent);
   for( i = 0; i < 3; i++)
      obs->obs_posn[i] += geocent[i];
   set_obs_vect( obs);
   obs->weight = weight;
   return( 0);
}

/*  The following does some simple error checking to 'buff' to see if it's
    really an observation,  or just a random line in the file.   */

static int is_an_observation( const char *buff, const char *obj_name)
{
   char tbuff[30];

   if( buff[44] == '-' || buff[44] == '+')
      if( strlen( buff) > 80 && (buff[12] == ' ' || buff[12] == '*'))
         if( (buff[15] == '1' || buff[15] == '2') && buff[19] == ' ')
            if( buff[25] == '.' && buff[40] == '.' && buff[14] != 's')
               if( buff[79] >= '0' && buff[79] <= '9')
                  {
                  get_object_name( tbuff, buff);
                  if( !obj_name || !strcmp( obj_name, tbuff))
                     return( 1);
                  }
   return( 0);
}

static int compare_observations( OBSERVE FAR *obs1, OBSERVE FAR *obs2)
{
   int rval;

   if( obs1->jd < obs2->jd)
      rval = -1;
   else if( obs1->jd > obs2->jd)
      rval = 1;
   else
      rval = FMEMCMP( obs1->mpc_code, obs2->mpc_code, 3);
   return( rval);
}

/* load_observations( ) reads in a file,  looking for observations of the
   object specified by 'obj_name'.  The return value is the array of
   observations,  with '*n_obs' containing the number of observations found.
   If 'station' is zero,  all observations are included.  Otherwise,  only
   those from that station are included... I used to have code that
   exploited that fact,  but right now,  'station' is always zero. */

OBSERVE FAR *load_observations( const char *filename, const char *obj_name,
                int *n_obs, int station)
{
   FILE *ifile = fopen( filename, "rb");
   char buff[100];
   OBSERVE FAR *rval;
   int i = 0, j;
   double weight = 1.;

   *n_obs = 0;
   if( !ifile)
      return( NULL);
   while( fgets( buff, sizeof( buff), ifile))
      if( is_an_observation( buff, obj_name))
         if( !station || atoi( buff + 76) == station)
            (*n_obs)++;
   rval = (OBSERVE FAR *)FCALLOC( *n_obs + 1, sizeof( OBSERVE));
   fseek( ifile, 0L, SEEK_SET);
   while( fgets( buff, sizeof( buff), ifile))
      {
      if( is_an_observation( buff, obj_name))
         if( !station || atoi( buff + 76) == station)
            if( parse_observation( rval + i++, buff, weight))
               {
               *n_obs = -1;
               fclose( ifile);
               return( NULL);
               }
            else if( buff[14] == 'S') /* satellite observation: second line */
               {
               OBSERVE FAR *tptr = rval + i - 1;
               double vect[3];
               int j;

               fgets( buff, sizeof( buff), ifile);
               for( j = 0; j < 3; j++)
                  {
                  vect[j] = atof( buff + 35 + j * 12);
                  if( buff[34 + j * 12] == '-')
                     vect[j] = -vect[j];
                  if( buff[32] == '1')   /* satellite obs offsets can be in */
                     vect[j] /= AU_IN_KM;   /* AU or,  less common,  km     */
                  }
               equatorial_to_ecliptic( vect);
               for( j = 0; j < 3; j++)
                  tptr->obs_posn[j] += vect[j];
               }
      if( !memcmp( buff, "#Weight ", 8))
         weight = atof( buff + 8);
      }
   fclose( ifile);
   for( i = 0; i < *n_obs; i++)
      for( j = i + 1; j < *n_obs; j++)
         if( compare_observations( rval + j, rval + i) < 0)
            {
            OBSERVE temp = rval[i];

            rval[i] = rval[j];
            rval[j] = temp;
            }

   for( i = 0; i < *n_obs - 1; i++)       /* search/destroy duplicate obs */
      if( !compare_observations( rval + i, rval + i + 1))
         {
         FMEMMOVE( rval + i, rval + i + 1, (*n_obs - i - 1) * sizeof( OBSERVE));
         (*n_obs)--;
         i--;
         }
   return( rval);
}

/* find_objects_in_file( ) reads through the file of MPC astrometric data
   specified by 'filename',  and figures out which objects appear in that
   file.  Those objects can then be listed on the console (DOS_FIND) or
   a scroll box (FIND_ORB),  and the user can select one for which she wants
   to determine an orbit. */

char *find_objects_in_file( const char *filename, int *n_found, int station)
{
   FILE *ifile = fopen( filename, "rb");
   char buff[100], *rval;
   int i, n = 0;

   if( !ifile)
      return( NULL);
   rval = (char *)calloc( 860, 1);
   while( fgets( buff, sizeof( buff), ifile))
      if( is_an_observation( buff, NULL))
         if( !station || atoi( buff + 76) == station)
            {
            if( *buff == '#')
               *buff = ' ';           /* handle remarked-out lines,  too */
            for( i = 0; i < n && memcmp( buff, rval + 16 * i, 12); i++)
               ;
            if( i == n)
               {
               if( n % 10 == 9)
                  rval = (char *)realloc( rval, 16 * (n + 11));
               memcpy( rval + 16 * n, buff, 14);
               *(short *)( rval + 16 * n + 14) = 0;
               n++;
               }
            (*(short *)( rval + 16 * i + 14))++;
            }
   *n_found = n;
   fclose( ifile);
   return( rval);
}


/* put_observer_data_in_text( ) takes a 'station_no' and fills 'buff'
   with a little bit of text about that station,  as found from
   STATIONS.TXT:  bits such as the lat/lon and name of the station.
   In FIND_ORB and DOS_FIND,  these details are shown for the station
   that collected the currently-selected observation. */

void put_observer_data_in_text( const char FAR *mpc_code, char *buff)
{
   if( get_observer_data( mpc_code, buff))
      {
      char tbuff[4];

      FMEMCPY( tbuff, mpc_code, 4);
      sprintf( buff, "No information about station '%s'", tbuff);
      }
   else
      {
      const double pi = 3.141592653589793238462643383279502884197;
      double lon = atof( buff + 4), lat, tan_u;
      double rho_cos_lat = atof( buff + 14);
      double rho_sin_lat = atof( buff + 23);
      int lat_sign = (buff[21] == '+' ? 'N' : 'S');
      int lon_sign = (lon > 180. ? 'W' : 'E');
      int i;

      memmove( buff, buff + 30, strlen( buff + 28));
      for( i = 0; buff[i] >= ' '; i++)
         ;
      if( rho_cos_lat)
         {
         tan_u = (rho_sin_lat / rho_cos_lat) / EARTH_AXIS_RATIO;
         lat = atan( tan_u / EARTH_AXIS_RATIO) * 180. / pi;
         if( lon > 180.)
            {
            lon = 360. - lon;
            lon_sign = 'W';
            }
         else
            lon_sign = 'E';
         sprintf( buff + i, "  (%c%.4lf %c%.4lf)", lat_sign, lat,
                                                   lon_sign, lon);
         }
      else
         strcpy( buff + i, "  (Geocentric)");
      }
}

const char *get_environment_ptr( const char *env_ptr)
{
   static char *edata;
   static unsigned edata_len = 0;
   unsigned i, j;
   const char *rval = NULL;
   const unsigned env_ptr_len = strlen( env_ptr);

   if( !edata)
      {
      FILE *ifile = fopen( "environ.dat", "rb");

      if( !ifile)
         {
         edata_len = 0;
         edata = "";
         }
      else
         {
         fseek( ifile, 0L, SEEK_END);
         edata_len = (unsigned)ftell( ifile);
         fseek( ifile, 0L, SEEK_SET);
         edata = (char *)malloc( edata_len);
         fread( edata, 1, edata_len, ifile);
         fclose( ifile);
         }
      for( i = 0; i < edata_len; i++)
         if( edata[i] == 10 || edata[i] == 13)
            edata[i] = 0;
      }

   i = 0;
   while( i < edata_len && !rval)
      {
      const char *tptr = edata + i;

      for( j = 0; tptr[j]; j++)
         ;
      if( j > env_ptr_len && !memcmp( tptr, env_ptr, env_ptr_len) &&
                         tptr[env_ptr_len] == '=')
         rval = tptr + env_ptr_len + 1;
      i += j;
      while( !tptr[j] && i < edata_len)
         {
         j++;
         i++;
         }
      }
   if( !rval)
      rval = "";
   return( rval);
}
