#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <malloc.h>
#include "watdefs.h"
#include "comets.h"
#include "mpc_obs.h"
#include "lsquare.h"
#include "date.h"

double stepsize = 3.;
int use_rk = 1;
int perturbers = 0;


#define PI 3.141592653589793238462643383279502884197
#define GAUSS_K .01720209895
#define SOLAR_GM (GAUSS_K * GAUSS_K)
#define ACCEL0 (GAUSS_K * GAUSS_K)
#define SQRT_ACCEL0 GAUSS_K
#define SQRT_2 1.41421356
#define AU_IN_KM 1.4959787e+8
#define AU_PER_DAY (86400. * 299792.456 / AU_IN_KM)
         /* ACCEL0 is in AU/day^2,  and is the acceleration at one AU */
         /* so that accel(r) = ACCEL0 / r^2 */


int initial_orbit( OBSERVE FAR *obs, int n_obs, double *orbit);
void integrate_orbit( double *orbit, double t0, double t1);
double compute_rms( const OBSERVE FAR *obs, int n_obs, int method);
int herget_method( OBSERVE FAR *obs, int n_obs, double r1, double r2,
                              double *orbit, double *d_r1, double *d_r2);
int adjust_herget_results( OBSERVE FAR *obs, int n_obs, double *orbit);
int set_locs( const double *orbit, double t0, OBSERVE FAR *obs, int n_obs);
double DLL_FUNC decimal_day_to_dmy( double jd, long *year, int *month);
int DLL_FUNC elements_in_mpc_format( char *obuff, const ELEMENTS *elem,
                const char *obj_id, const int is_cometary, const int format);
double calc_obs_magnitude( int is_comet, double obj_sun, double obj_earth,
                                double earth_sun);  /* orb_func.cpp */
int get_r1_and_r2( const int n_obs, const OBSERVE FAR *obs,
                             double *r1, double *r2);    /* orb_func.cpp */
int find_best_fit_planet( const double jd, const double *ivect,
                                 double *rel_vect, const int included);
void observation_summary_data( char *obuff, const OBSERVE FAR *obs,
                                const int n_obs);        /* orb_func.cpp */
double calc_absolute_magnitude( const OBSERVE FAR *obs, int n_obs);
void full_improvement( OBSERVE FAR *obs, int n_obs, double *orbit,
               double epoch, const char *limited_orbit);    /* orb_func.cpp */
const char *get_environment_ptr( const char *env_ptr);     /* mpc_obs.cpp */

#define CACHE_NOT_IN_USE 0
#define CACHE_GROWING 1
#define CACHE_USING 2

         /* The 32-bit compilers allows for a much larger cache of        */
         /* precomputed planetary positions than the 16-bit compilers do: */
#ifdef BITS_32
#define CACHE_SIZE 160000
#else
#define CACHE_SIZE (60000u / 24u)
#endif

int reset_cache( int action, int size);

static void set_distance( OBSERVE FAR *obs, double r)
{
   int i;
   double d = 0.;

   obs->r = r;
   for( i = 0; i < 3; i++)
      {
      double p = obs->obs_posn[i] + r * obs->vect[i];

      obs->obj_posn[i] = p;
      d += p * p;
      }
   obs->solar_r = sqrt( d);
}

static double find_r_given_solar_r( const OBSERVE FAR *obs,
                                                     const double solar_r)
{
   double r_dot_v = 0., r_dot_r = 0., b, c, discr, rval = -1.;
   int i;

   for( i = 0; i < 3; i++)
      {
      r_dot_r += obs->obs_posn[i] * obs->obs_posn[i];
      r_dot_v += obs->obs_posn[i] * obs->vect[i];
      }
   b = 2. * r_dot_v;
   c = r_dot_r - solar_r * solar_r;
   discr = b * b - 4 * c;
   if( discr > 0.)
      rval = (-b + sqrt( discr)) / 2.;
   return( rval);
}

int full_rk_step( double *ivals, double *ovals, double t0, double t1,
                              int n_vals, double max_err);
int calc_derivatives( double jd, double *ival, double *oval);

void integrate_orbit( double *orbit, double t0, double t1)
{
   while( t0 != t1)
      {
      double delta_t, r2 = 0, afact;
      int i;

      if( t1 > t0)
         delta_t = ((stepsize > t1 - t0) ? t1 - t0 : stepsize);
      else
         delta_t = ((stepsize > t0 - t1) ? t1 - t0 : -stepsize);

      if( use_rk)
         full_rk_step( orbit, orbit, t0, t0 + delta_t, 6, 1.e-10);
      else
         if( perturbers)
            {
            double deriv[6];

            calc_derivatives( t0, orbit, deriv);
            for( i = 0; i < 6; i++)
               orbit[i] += delta_t * deriv[i];
            }
         else
            {
            for( i = 0; i < 3; i++)
               r2 += orbit[i] * orbit[i];
            afact = -ACCEL0 * delta_t / (sqrt( r2) * r2);

            for( i = 0; i < 3; i++)
               {
               double aval = afact * orbit[i];

               orbit[i] += delta_t * orbit[i + 3];
               orbit[i + 3] += aval;
               }
            }
      t0 += delta_t;
      }
}

static void light_time_lag( const double *orbit, const double *observer, double *result)
{
   double r = 0., delta, dt, solar_r2 = 0., afact;
   int i;

   for( i = 0; i < 3; i++)
      {
      delta = orbit[i] - observer[i];
      r += delta * delta;
      solar_r2 += orbit[i] * orbit[i];
      }
   r = sqrt( r);
   dt = -r / AU_PER_DAY;
   afact = 1. - ACCEL0 * .5 * dt * dt / (sqrt( solar_r2) * solar_r2);
   for( i = 0; i < 3; i++)
      result[i] = afact * orbit[i] + dt * orbit[i + 3];
}

static void set_solar_r( OBSERVE FAR *ob)
{
   double r = 0.;
   int i;

   for( i = 0; i < 3; i++)
      r += ob->obj_posn[i] * ob->obj_posn[i];
   ob->solar_r = sqrt( r);
}

int set_locs( const double *orbit, double t0, OBSERVE FAR *obs, int n_obs)
{
   int i, j, *idx = (int *)calloc( n_obs, sizeof( int));
   double curr_orbit[6], curr_t, tvals[6];

   for( i = 0; i < n_obs; i++)
      idx[i] = i;
   for( i = 0; i < n_obs; i++)
      for( j = i + 1; j < n_obs; j++)
         if( obs[idx[i]].jd > obs[idx[j]].jd)
            {
            int temp = idx[i];
            idx[i] = idx[j];
            idx[j] = temp;
            }

   for( i = 0; i < n_obs && obs[idx[i]].jd < t0; i++)
      ;

               /* integrate _forward_... */
   memcpy( curr_orbit, orbit, 6 * sizeof( double));
   curr_t = t0;
   for( j = i; j < n_obs; j++)
      {
      OBSERVE FAR *optr = obs + idx[j];

      integrate_orbit( curr_orbit, curr_t, optr->jd);
      curr_t = optr->jd;
      FMEMCPY( tvals, optr->obs_posn, 3 * sizeof( double));
      light_time_lag( curr_orbit, tvals, tvals + 3);
      FMEMCPY( optr->obj_posn, tvals + 3, 3 * sizeof( double));
      }

               /* ...then integrate _backward_... */
   memcpy( curr_orbit, orbit, 6 * sizeof( double));
   curr_t = t0;
   for( j = i - 1; j >= 0; j--)
      {
      OBSERVE FAR *optr = obs + idx[j];

      integrate_orbit( curr_orbit, curr_t, optr->jd);
      curr_t = optr->jd;
      FMEMCPY( tvals, optr->obs_posn, 3 * sizeof( double));
      light_time_lag( curr_orbit, tvals, tvals + 3);
      FMEMCPY( optr->obj_posn, tvals + 3, 3 * sizeof( double));
      }

   free( idx);

   for( i = 0; i < n_obs; i++)
      {
      double loc[3], ra, dec, temp, r = 0.;
      static const double sin_obliq_2000 = .397777156;
      static const double cos_obliq_2000 = .917482062;

      for( j = 0; j < 3; j++)
         {
         loc[j] = obs[i].obj_posn[j] - obs[i].obs_posn[j];
         r += loc[j] * loc[j];
         }
      r = sqrt( r);
      obs[i].r = r;
      temp = loc[1] * cos_obliq_2000 - loc[2] * sin_obliq_2000;
      loc[2] = loc[2] * cos_obliq_2000 + loc[1] * sin_obliq_2000;
      loc[1] = temp;
      ra = atan2( loc[1], loc[0]);
      dec = asin( loc[2] / r);
      while( ra - obs[i].ra > PI)
         ra -= 2. * PI;
      while( ra - obs[i].ra < -PI)
         ra += 2. * PI;
      obs[i].computed_ra = ra;
      obs[i].computed_dec = dec;
      set_solar_r( obs + i);
      }
   return( 0);
}

double compute_rms( const OBSERVE FAR *obs, int n_obs, int method)
{
   double rval = 0., d2;
   int i, n_included = 0;

   for( i = n_obs; i; i--, obs++)
      if( obs->is_included)
         {
         double d_ra  = obs->computed_ra  - obs->ra;
         double d_dec = obs->computed_dec - obs->dec;

         d_ra *= cos( d_dec);
         d2 = d_dec * d_dec + d_ra * d_ra;
         if( method == 1)
            rval += d2;
         if( method == 2)
            rval += sqrt( d2);
         n_included++;
         }
   rval /= (double)n_included;
   if( method == 1)
      rval = sqrt( rval);
   return( 3600. * (180. / PI) * rval);
}

static double eval_3x3_determinant( const double *a, const double *b,
                         const double *c)
{
   return( a[0] * (b[1] * c[2] - c[1] * b[2])
         + b[0] * (c[1] * a[2] - a[1] * c[2])
         + c[0] * (a[1] * b[2] - b[1] * a[2]));
}

static int find_transfer_orbit( double *orbit, OBSERVE FAR *obs1,
                                    OBSERVE FAR *obs2)
{
   double r = 0., delta_t = obs2->jd - obs1->jd;
   double deriv[6];
   double diff_squared = 999., accel0;
   int i, max_iterations = 10;

   set_distance( obs1, obs1->r);
   set_distance( obs2, obs2->r);

   r = sqrt( obs1->solar_r * obs2->solar_r);
   accel0 = -ACCEL0 / (r * r * r);
   for( i = 0; i < 3; i++)
      {
                  /* find midpoint between obs1 and obs2... */
      orbit[i] = (obs2->obj_posn[i] + obs1->obj_posn[i]) / 2.;
                  /* and the average velocity connecting them: */
      orbit[i + 3] = (obs2->obj_posn[i] - obs1->obj_posn[i]) / delta_t;
      r += orbit[i] * orbit[i];
      }

   calc_derivatives( (obs1->jd + obs2->jd) / 2., orbit, deriv);
   for( i = 0; i < 3; i++)
      {
      orbit[i] = obs1->obj_posn[i];
      orbit[3 + i] -= deriv[i + 3] * delta_t / 2.;
      }

   reset_cache( CACHE_GROWING, CACHE_SIZE);
            /* Iterate until the error is less than 1e-8 of the object */
            /* distance.  This corresponds to an error of about .002 arcsec */
            /* in the actual position... */
            /*   Assume a maximum of five iterations,  just to be safe */
   while( sqrt( diff_squared) > 1.e-8 * obs2->r && max_iterations--)
      {
      double delta[4][3], discr;
      int pass;

      for( pass = 0; pass < 4; pass++)
         {
         double orbit2[6];
         const double h = 1.e-5;

         memcpy( orbit2, orbit, 6 * sizeof( double));
         if( pass)
            orbit2[(pass - 1) + 3] += h;
         integrate_orbit( orbit2, obs1->jd - obs1->r / AU_PER_DAY,
                                  obs2->jd - obs2->r / AU_PER_DAY);
         for( i = 0; i < 3; i++)
            orbit2[i] -= obs2->obj_posn[i];
         memcpy( delta[pass], orbit2, 3 * sizeof( double));
         if( pass)
            for( i = 0; i < 3; i++)
               delta[pass][i] = (delta[pass][i] - delta[0][i]) / h;
         reset_cache( CACHE_USING, 0);
         }
      discr = eval_3x3_determinant( delta[1], delta[2], delta[3]);
      orbit[3] -= eval_3x3_determinant( delta[0], delta[2], delta[3]) / discr;
      orbit[4] -= eval_3x3_determinant( delta[1], delta[0], delta[3]) / discr;
      orbit[5] -= eval_3x3_determinant( delta[1], delta[2], delta[0]) / discr;
      diff_squared = 0.;
      for( i = 0; i < 3; i++)
         diff_squared += delta[0][i] * delta[0][i];
#ifdef CURRENTLY_OBSOLETE_DEBUGGING_CODE
         {
         FILE *ofile = fopen( "debug.dat", "ab");

         fprintf( ofile, "%d: %lf (%lf %lf %lf)\n",
                       10 - max_iterations, sqrt( diff_squared),
                       delta[0][0], delta[0][1], delta[0][2]);
         fclose( ofile);
         }
#endif
      }
   reset_cache( CACHE_NOT_IN_USE, 0);
               /* adjust for light-time lag: */
   r = -ACCEL0 / (obs1->solar_r * obs1->solar_r * obs1->solar_r);
   for( i = 0; i < 3; i++)
      {
      double time_lag = obs1->r / AU_PER_DAY;

      orbit[i] += time_lag * (orbit[3 + i] + orbit[i] * r * time_lag / 2.);
      orbit[i + 3] += orbit[i] * r * time_lag;
      }
   return( 0);
}

#define RAD2SEC (180. * 3600. / PI)

void set_obs_vect( OBSERVE FAR *obs);        /* mpc_obs.h */

int adjust_herget_results( OBSERVE FAR *obs, int n_obs, double *orbit)
{
   int i, n_found = 0, rval = -1;
   double avg_x2 = 0., avg_x = 0., avg_xt = 0., avg_t = 0.;
   double avg_y2 = 0., avg_y = 0., avg_yt = 0., avg_t2 = 0.;
// FILE *debug_file = fopen( "debug.dat", "wb");

   while( n_obs && !obs->is_included)
      {
      obs++;
      n_obs--;
      }
   while( n_obs && !obs[n_obs - 1].is_included)
      n_obs--;

   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         double dx = obs[i].computed_ra - obs[i].ra;
         double dy = obs[i].computed_dec - obs[i].dec;
         double dt = obs[i].jd - obs[0].jd;

         avg_x2 += dx * dx;
         avg_y2 += dy * dy;
         avg_xt += dx * dt;
         avg_yt += dy * dt;
         avg_x += dx;
         avg_y += dy;
         avg_t += dt;
         avg_t2 += dt * dt;
         n_found++;
         }

// fprintf( debug_file, "%d found\n", n_found);
   if( n_found)
      {
      double determ;

      avg_x2 /= (double)n_found;
      avg_y2 /= (double)n_found;
      avg_xt /= (double)n_found;
      avg_yt /= (double)n_found;
      avg_x /= (double)n_found;
      avg_y /= (double)n_found;
      avg_t /= (double)n_found;
      avg_t2 /= (double)n_found;
//    fprintf( debug_file, "avg_x2 = %lf   avg_y2 = %lf\n",
//                avg_x2 * RAD2SEC * RAD2SEC, avg_y2 * RAD2SEC * RAD2SEC);
//    fprintf( debug_file, "avg_x = %lf   avg_y = %lf\n",
//                avg_x * RAD2SEC, avg_y * RAD2SEC);
      determ = avg_t2 - avg_t * avg_t;
      if( determ)
         {
         double ax = (avg_xt - avg_x * avg_t) / determ;
         double bx = (avg_t2 * avg_x - avg_t * avg_xt) / determ;
         double ay = (avg_yt - avg_y * avg_t) / determ;
         double by = (avg_t2 * avg_y - avg_t * avg_yt) / determ;
         OBSERVE start = obs[0];
         OBSERVE end = obs[n_obs - 1];

         start.ra  = start.computed_ra - bx;
         end.ra    = end.computed_ra - (bx + (end.jd - start.jd) * ax);
         start.dec = start.computed_dec - by;
         end.dec   = end.computed_dec - (by + (end.jd - start.jd) * ay);
//       fprintf( debug_file, "Shift at start: %lf, %lf\n",
//                bx * RAD2SEC, by * RAD2SEC);
//       fprintf( debug_file, "Shift at end: %lf, %lf\n",
//                (bx + (end.jd - start.jd) * ax) * RAD2SEC,
//                (by + (end.jd - start.jd) * ay) * RAD2SEC);
         set_obs_vect( &start);
         set_obs_vect( &end);
         set_distance( &start, start.r);
         set_distance( &end, end.r);
         find_transfer_orbit( orbit, &start, &end);
         set_locs( orbit, start.jd, obs, n_obs);
         rval = 0;
         }
      }
// fclose( debug_file);
   return( rval);
}

         /* Don't attempt to use the Herget method over a time span  */
         /* greater than 300 days... a somewhat arbitrary limit;  it */
         /* really ought to be lower for orbits closer to the sun,   */
         /* larger for Centaur/TNO-type objects                      */

#define MAX_HERGET_TIMESPAN 300

int herget_method( OBSERVE FAR *obs, int n_obs, double r1, double r2,
                              double *orbit, double *d_r1, double *d_r2)
{
   double *xresid, *yresid, *dx_dr1, *dx_dr2, *dy_dr1, *dy_dr2;
   double delta, a = 0., b = 0., c = 0., e = 0., f = 0., determ;
   int i, n_real_obs = 0, using_pseudo_vaisala = (r1 < 0.);
   double orbit2[6];

   for( i = 0; i < n_obs; i++)
      {
      obs[i].computed_ra  = obs[i].ra;
      obs[i].computed_dec = obs[i].dec;
      if( obs[i].is_included)
         n_real_obs++;
      }

   while( n_obs && !obs->is_included)
      {
      obs++;
      n_obs--;
      }
   while( n_obs && (!obs[n_obs - 1].is_included ||
            obs[n_obs - 1].jd - obs[0].jd > MAX_HERGET_TIMESPAN))
      {
      n_obs--;
      obs[n_obs].is_included = 0;
      }

   if( n_real_obs < 2)
      return( -1);

   if( using_pseudo_vaisala)
      {
      r2 = find_r_given_solar_r( obs + n_obs - 1, -r1);
      r1 = find_r_given_solar_r( obs, -r1);
      if( d_r1)
         *d_r1 = r1;
      if( d_r2)
         *d_r1 = r2;
      }
   set_distance( obs, r1);
   set_distance( obs + n_obs - 1, r2);

   find_transfer_orbit( orbit, obs, obs + n_obs - 1);
   set_locs( orbit, obs[0].jd, obs, n_obs);
   if( !d_r1 || !d_r2 || using_pseudo_vaisala)
      return( 0);

   *d_r1 = *d_r2 = 0.;
   if( n_real_obs == 2)
      return( 0);       /* good as done... */

   xresid = (double *)calloc( 6 * n_obs, sizeof( double));
   if( !xresid)
      return( -1);
   yresid = xresid + n_obs;
   dx_dr1 = xresid + n_obs * 2;
   dx_dr2 = xresid + n_obs * 3;
   dy_dr1 = xresid + n_obs * 4;
   dy_dr2 = xresid + n_obs * 5;

   for( i = 0; i < n_obs; i++)
      {
      xresid[i] = obs[i].computed_ra - obs[i].ra;
      yresid[i] = obs[i].computed_dec - obs[i].dec;
      }

   delta = r1 / 10000.;
   if( delta > .1) delta = .1;
   set_distance( obs, r1 + delta);
   find_transfer_orbit( orbit2, obs, obs + n_obs - 1);
   set_locs( orbit2, obs[0].jd, obs, n_obs);
   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         double dx = obs[i].computed_ra - obs[i].ra;
         double dy = obs[i].computed_dec - obs[i].dec;

         dx_dr1[i] = (dx - xresid[i]) / delta;
         dy_dr1[i] = (dy - yresid[i]) / delta;
         }

   delta = r2 / 10000.;
   if( delta > .1) delta = .1;
   set_distance( obs, r1);
   set_distance( obs + n_obs - 1, r2 + delta);
   find_transfer_orbit( orbit2, obs, obs + n_obs - 1);
   set_locs( orbit2, obs[0].jd, obs, n_obs);
   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         double dx = obs[i].computed_ra - obs[i].ra;
         double dy = obs[i].computed_dec - obs[i].dec;

         dx_dr2[i] = (dx - xresid[i]) / delta;
         dy_dr2[i] = (dy - yresid[i]) / delta;
         }
                    /* OK,  we now have all values & derivatives needed... */
   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         c += xresid[i] * dx_dr1[i] + yresid[i] * dy_dr1[i];
         a += dx_dr1[i] * dx_dr1[i] + dy_dr1[i] * dy_dr1[i];
         b += dx_dr1[i] * dx_dr2[i] + dy_dr1[i] * dy_dr2[i];

         f += xresid[i] * dx_dr2[i] + yresid[i] * dy_dr2[i];
         /*  d = b;  */
         e += dx_dr2[i] * dx_dr2[i] + dy_dr2[i] * dy_dr2[i];
         }

   free( xresid);
   determ = a * e - b * b;
   *d_r1 = -(e * c - b * f) / determ;
   *d_r2 = -(a * f - c * b) / determ;
   if( *d_r1 > r1 / 3.) *d_r1 = r1 / 3.;
   if( *d_r1 <-r1 / 3.) *d_r1 = -r1 / 3.;
   if( *d_r2 > r2 / 3.) *d_r2 = r2 / 3.;
   if( *d_r2 <-r2 / 3.) *d_r2 = -r2 / 3.;
   return( 0);
}

static int setup_parabolic( const double *iparams, double *orbit)
{
   double r = 0., escape_vel;
   int i;

   for( i = 0; i < 3; i++)
      {
      orbit[i] = iparams[i];
      r += iparams[i] * iparams[i];
      }
   r = sqrt( r);
   escape_vel = SQRT_2 * SQRT_ACCEL0 / sqrt( r);
   orbit[3] = escape_vel * cos( iparams[3]) * cos( iparams[4]);
   orbit[4] = escape_vel * sin( iparams[3]) * cos( iparams[4]);
   orbit[5] = escape_vel *                    sin( iparams[4]);
   return( 0);
}

void improve_parabolic( OBSERVE FAR *obs, int n_obs, double *orbit,
                                                              double epoch)
{
   void *lsquare = lsquare_init( 5);
   double *xresids = (double *)calloc( 2 * n_obs + 10 * n_obs, sizeof( double));
   double *yresids = xresids + n_obs;
   double *slopes = yresids + n_obs;
   double params2[5], params[5], differences[5];
   const double delta_val = .0000001;
   double v2 = 0.;
   int i, j;

   memcpy( params, orbit, 3 * sizeof( double));
   params[3] = atan2( orbit[4], orbit[3]);
   for( i = 3; i < 6; i++)
      v2 += orbit[i] * orbit[i];
   params[4] = asin( orbit[5] / sqrt( v2));
   setup_parabolic( params, orbit);

   reset_cache( CACHE_GROWING, CACHE_SIZE);

   set_locs( orbit, epoch, obs, n_obs);
   for( i = 0; i < n_obs; i++)
      {
      xresids[i] = obs[i].computed_ra - obs[i].ra;
      yresids[i] = obs[i].computed_dec - obs[i].dec;
      }

   for( i = 0; i < 5; i++)
      {
      memcpy( params2, params, 5 * sizeof( double));
      params2[i] = params[i] - delta_val;
      setup_parabolic( params2, orbit);
      reset_cache( CACHE_USING, 0);
      set_locs( orbit, epoch, obs, n_obs);
      for( j = 0; j < n_obs; j++)
         {
         slopes[i + j * 10] = obs[j].computed_ra;
         slopes[i + j * 10 + 5] = obs[j].computed_dec;
         }

      params2[i] = params[i] + delta_val;
      setup_parabolic( params2, orbit);
      reset_cache( CACHE_USING, 0);
      set_locs( orbit, epoch, obs, n_obs);
      for( j = 0; j < n_obs; j++)
         {
         slopes[i + j * 10] -= obs[j].computed_ra;
         slopes[i + j * 10 + 5] -= obs[j].computed_dec;
         slopes[i + j * 10] /= 2. * delta_val;
         slopes[i + j * 10 + 5] /= 2. * delta_val;
         }
      }

   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         lsquare_add_observation( lsquare, xresids[i], 1., slopes + i * 10);
         lsquare_add_observation( lsquare, yresids[i], 1., slopes + i * 10 + 5);
         }

   free( xresids);
   lsquare_solve( lsquare, differences);
   lsquare_free( lsquare);

   for( i = 0; i < 5; i++)
      params[i] += differences[i];
   setup_parabolic( params, orbit);
   reset_cache( CACHE_USING, 0);
   set_locs( orbit, epoch, obs, n_obs);
   reset_cache( CACHE_NOT_IN_USE, 0);
}

#ifdef OLD_DEBUGGING_CODE
#include <windows.h>
#endif

static int evaluate_limited_orbit( const double *orbit,
                    const int planet_orbiting, const double epoch,
                    const char *limited_orbit, double *constraints)
{
   int rval = 0;

   if( limited_orbit)
      {
      ELEMENTS elem;
      extern double planet_mass[];

      calc_classical_elements( &elem, orbit, epoch, 1,
                              SOLAR_GM * planet_mass[planet_orbiting]);
      while( *limited_orbit && limited_orbit[1] == '=')
         {
         double value = atof( limited_orbit + 2);
         char variable = *limited_orbit;

         while( *limited_orbit && *limited_orbit != ',')
            limited_orbit++;
         switch( variable)
            {
            case 'q':
               constraints[rval++] = elem.major_axis * (1. - elem.ecc) - value;
               break;
            case 'Q':
               constraints[rval++] = elem.major_axis * (1. + elem.ecc) - value;
               break;
            case 'a':
               constraints[rval++] = elem.major_axis - value;
               break;
            case 'e':
               if( value)
                  constraints[rval++] = elem.ecc - value;
               else        /* handle e=0 (circular) orbits separately: */
                  {
                  constraints[rval++] = orbit[0] * orbit[3] +
                              orbit[1] * orbit[4] + orbit[2] * orbit[5];
                  constraints[rval++] = sqrt( orbit[0] * orbit[0] +
                               orbit[1] * orbit[1] + orbit[2] * orbit[2])
                               - elem.major_axis;
                  }
               break;
            case 'i':
               constraints[rval++] = elem.incl * 180. / PI - value;
               break;
            case 'P':
               {
               double period = elem.major_axis *
                        sqrt( elem.major_axis / planet_mass[planet_orbiting]);

               if( limited_orbit[-1] == 'd')       /* cvt to days */
                  period *= 365.25;
               if( limited_orbit[-1] == 'm')       /* minutes */
                  period *= 365.25 * 1440.;

               constraints[rval++] = period - value;
               }
               break;
            }
         if( *limited_orbit == ',')
            limited_orbit++;
         }
      }
   return( rval);
}

#define MAX_CONSTRAINTS 5

void full_improvement( OBSERVE FAR *obs, int n_obs, double *orbit,
                  double epoch, const char *limited_orbit)
{
   void *lsquare = lsquare_init( 6);
   double FAR *xresids;
   double FAR *yresids;
   double FAR *slopes;
   double constraint_slope[MAX_CONSTRAINTS * 6];
   double orbit2[6], differences[6], real_epoch = epoch;
   double orbit_offset[6];
   double constraint[MAX_CONSTRAINTS];
   int planet_orbiting, n_constraints = 0;
   int i, j;
   const char *covariance_filename = get_environment_ptr( "COVARIANCE");

   for( i = 0; i < n_obs; i++)
      {
      obs[i].computed_ra  = obs[i].ra;
      obs[i].computed_dec = obs[i].dec;
      }

   while( n_obs && !obs->is_included)
      {
      obs++;
      n_obs--;
      }
   while( n_obs && !obs[n_obs - 1].is_included)
      n_obs--;
   if( n_obs < 3)
      ;

   real_epoch = epoch;
// if( obs->jd > epoch)
//    real_epoch = obs->jd;
// else if( obs[n_obs - 1].jd < epoch)
//    real_epoch = obs[n_obs - 1].jd;

   if( real_epoch != epoch)
      integrate_orbit( orbit, epoch, real_epoch);

   if( limited_orbit)
      {
      planet_orbiting = find_best_fit_planet( real_epoch, orbit, orbit2, -1);
      for( i = 0; i < 6; i++)
         orbit_offset[i] = orbit[i] - orbit2[i];
      n_constraints = evaluate_limited_orbit( orbit2, planet_orbiting,
                                       real_epoch, limited_orbit, constraint);
      }

   xresids = (double FAR *)FCALLOC( 14 * n_obs + 6, sizeof( double));
   yresids = xresids + n_obs;
   slopes = yresids + n_obs;

   reset_cache( CACHE_GROWING, CACHE_SIZE);
   set_locs( orbit, real_epoch, obs, n_obs);
   for( i = 0; i < n_obs; i++)
      {
      xresids[i] = obs[i].computed_ra - obs[i].ra;
      yresids[i] = obs[i].computed_dec - obs[i].dec;
      }

   for( i = 0; i < 6; i++)
      {
      double delta_val = (i < 3 ? 1.e-8 : 1.e-11);
//    double delta_val = (i < 3 ? 1.e-6 : 1.e-9);

      memcpy( orbit2, orbit, 6 * sizeof( double));
      orbit2[i] = orbit[i] - delta_val;
      reset_cache( CACHE_USING, 0);
      set_locs( orbit2, real_epoch, obs, n_obs);
      for( j = 0; j < n_obs; j++)
         {
         slopes[i + j * 12] = obs[j].computed_ra;
         slopes[i + j * 12 + 6] = obs[j].computed_dec;
         }

      orbit2[i] = orbit[i] + delta_val;
      reset_cache( CACHE_USING, 0);
      set_locs( orbit2, real_epoch, obs, n_obs);
      for( j = 0; j < n_obs; j++)
         {
         slopes[i + j * 12] -= obs[j].computed_ra;
         slopes[i + j * 12 + 6] -= obs[j].computed_dec;
         slopes[i + j * 12] /= 2. * delta_val;
         slopes[i + j * 12 + 6] /= 2. * delta_val;
         }

      if( limited_orbit)
         {
         double constraint2[MAX_CONSTRAINTS];

         for( j = 0; j < 6; j++)
            orbit2[j] -= orbit_offset[j];
         evaluate_limited_orbit( orbit2, planet_orbiting, real_epoch,
                                   limited_orbit, constraint2);
         for( j = 0; j < n_constraints; j++)
            constraint_slope[i + j * 6] =
                     (constraint[j] - constraint2[j]) / delta_val;
         }
      }

   for( i = 0; i < n_obs; i++)
      if( obs[i].is_included)
         {
         double loc_vals[12];

         FMEMCPY( loc_vals, slopes + i * 12, 12 * sizeof( double));
         lsquare_add_observation( lsquare, xresids[i] * obs[i].weight, 1., loc_vals);
         lsquare_add_observation( lsquare, yresids[i] * obs[i].weight, 1., loc_vals + 6);
         }

   if( limited_orbit)
      for( j = 0; j < n_constraints; j++)
         lsquare_add_observation( lsquare, constraint[j], 1.,
                                            constraint_slope + j * 6);

   FFREE( xresids);
   if( *covariance_filename)
      {
      FILE *ofile = fopen( covariance_filename, "wb");
      double *matrix = lsquare_covariance_matrix( lsquare);

      fprintf( ofile, "Covariance:\n");
      for( i = 0; i < 36; i++)
         fprintf( ofile, "%.5lg%c", matrix[i], (i % 6 == 5 ? '\n' : ' '));

      fprintf( ofile, "Correlation:\n");
      for( i = 0; i < 6; i++)
         for( j = 0; j < 6; j++)
            {
            double oval =
                    matrix[i + j * 6] / sqrt( matrix[i * 7] * matrix[j * 7]);
            fprintf( ofile, "%.8lf%c", oval, (j == 5 ? '\n' : ' '));
            }
      fclose( ofile);
      free( matrix);
      }
   lsquare_solve( lsquare, differences);
   lsquare_free( lsquare);
   for( i = 0; i < 6; i++)
      {
      if( differences[i] > .1)
         differences[i] = .1;
      if( differences[i] < -.1)
         differences[i] = -.1;
      orbit[i] += differences[i];
      }

   reset_cache( CACHE_USING, 0);
   set_locs( orbit, real_epoch, obs, n_obs);
   reset_cache( CACHE_NOT_IN_USE, 0);

   if( real_epoch != epoch)
      integrate_orbit( orbit, real_epoch, epoch);
}

#define MAX_INITIAL_SPAN 150.

static double score_orbit_arc( const OBSERVE FAR *obs, const int n_obs)
{
   double longest_omission = 0.;
   double arc_length = obs[n_obs - 1].jd - obs[0].jd;
   int i, j;

   for( i = 0; i < n_obs - 1; i++)
      if( obs[i].is_included)
         {
         for( j = i + 1; j < n_obs && !obs[j].is_included; j++)
            ;
         if( j < n_obs)
            if( longest_omission < obs[j].jd - obs[i].jd)
               longest_omission = obs[j].jd - obs[i].jd;
         }
   return( arc_length - longest_omission);
}

static double evaluate_initial_orbit( const OBSERVE FAR *obs, const int n_obs,
                                                      const double *orbit)
{
   extern double planet_mass[];
   double rms_err = compute_rms( obs, n_obs, 1), rval, rel_orbit[6];
   ELEMENTS elem;
   int planet_orbiting = find_best_fit_planet( obs->jd,
                                  orbit, rel_orbit, perturbers | 1);

   calc_classical_elements( &elem, rel_orbit, obs[0].jd, 1,
                              SOLAR_GM * planet_mass[planet_orbiting]);
   rval = rms_err + elem.ecc;
   if( elem.ecc > .5)                     /* discourage eccentric orbits */
      rval += (elem.ecc - .5) * 2.;
   if( elem.ecc > 1.01)                /* _strongly_ discourage hyperbolics */
      rval += (elem.ecc - 1.01) * 1000.;
   return( rval);
}

int initial_orbit( OBSERVE FAR *obs, int n_obs, double *orbit)
{
   int i, j, start, end, quit, rval = -3;
   int old_perturbers;
   double pseudo_r, best_score = 0., score;
   FILE *start_ifile;

   for( i = 0; i < n_obs; i++)
      {
      obs[i].computed_ra  = obs[i].ra;
      obs[i].computed_dec = obs[i].dec;
      }

   while( n_obs && !obs->is_included)
      {
      obs++;
      n_obs--;
      }
   while( n_obs && !obs[n_obs - 1].is_included)
      n_obs--;
   if( n_obs < 2)
      return( -1);
   if( obs[n_obs - 1].jd - obs[0].jd < MAX_INITIAL_SPAN)
      {
      start = 0;
      end = n_obs - 1;
      }
   else           /* hunt for the best 'sub-arc' in the observations */
      {
      for( i = 0; i < n_obs - 1; i++)
         if( obs[i].is_included)
            for( j = i + 1; j < n_obs &&
                      obs[j].jd - obs[i].jd < MAX_INITIAL_SPAN; j++)
               if( obs[j].is_included)
                  {
                  score = score_orbit_arc( obs + i, j - i + 1);
                  if( score > best_score)
                     {
                     start = i;
                     end = j;
                     best_score = score;
                     }
                  }
      if( !best_score)   /* didn't find _any_ suitable arc */
         return( -2);
      for( i = 0; i < start; i++)
         obs[i].is_included = 0;
      for( i = end + 1; i < n_obs; i++)
         obs[i].is_included = 0;
      obs += start;
      n_obs = end - start + 1;
      }
   best_score = 1e+7;
   quit = 0;
   old_perturbers = perturbers;      /* temporarily suppress perturbations */
   start_ifile = fopen( "iparams.dat", "rb");
   if( !start_ifile)                /* default:  do things 'normally' */
      {
      perturbers = 0;
      for( i = 0; i < 2 && !quit; i++)
         for( pseudo_r = .01; !quit && pseudo_r < (i ? 5. : 100.);
                                                         pseudo_r *= 1.2)
            {
            double temp_orbit[6], pseudo_r_to_use;

            if( i)                          /* 2nd pass, dist from observer */
               pseudo_r_to_use = pseudo_r;
            else                            /* 1st pass, dist from sun */
               pseudo_r_to_use = -(1. + pseudo_r);
            herget_method( obs, n_obs, pseudo_r_to_use, pseudo_r_to_use,
                                temp_orbit, NULL, NULL);
            score = evaluate_initial_orbit( obs, n_obs, temp_orbit);
            if( score < best_score)
               {
               memcpy( orbit, temp_orbit, 6 * sizeof( double));
               rval = 0;
               best_score = score;
               }
            }
      }
   else   /* a 'start_ifile' was opened: handling a satellite */
      {
      double r1_center, r2_center, range, subrange;
      int n_steps, n_substeps, i, j, planet_no;

      fscanf( start_ifile, "%x %lf %lf %lf %lf %d %d",
            &planet_no, &r1_center, &r2_center, &range, &subrange,
                               &n_steps, &n_substeps);
      perturbers = (1 << planet_no);
      fclose( start_ifile);

      for( i = -n_steps; i <= n_steps; i++)
         for( j = -n_substeps; j <= n_substeps; j++)
            {
            double temp_orbit[6];

            herget_method( obs, n_obs,
                       r1_center + (double)i * range / (double)n_steps,
                       r2_center + (double)i * range / (double)n_steps
                                 + (double)j * subrange / (double)n_substeps,
                       temp_orbit, NULL, NULL);
            score = evaluate_initial_orbit( obs, n_obs, temp_orbit);
            if( score < best_score)
               {
               memcpy( orbit, temp_orbit, 6 * sizeof( double));
               rval = 0;
               best_score = score;
               }
            }
      }

   set_locs( orbit, obs[0].jd, obs, n_obs);
// perturbers = old_perturbers;        /* restore perturbers, if any */
   return( rval);
}

/* If you've got n_obs observations stored in the obs array,  the
   get_r1_and_r2( ) function will puzzle through them to find the first
   and last valid observation (those that haven't had their 'is_included'
   flags set to FALSE),  and will store the distances to them in *r1 and
   *r2.  These are shown near the top of the display,  and are used in
   the method of Herget.   */

int get_r1_and_r2( const int n_obs, const OBSERVE FAR *obs,
                                             double *r1, double *r2)
{
   int i, rval;

   for( i = 0; i < n_obs && (!obs[i].is_included || !obs[i].r); i++)
      ;
   if( i == n_obs)
      {
      *r1 = *r2 = 0.;
      rval = -1;
      }
   else
      {
      *r1 = obs[i].r;
      for( i = n_obs - 1; i && (!obs[i].is_included || !obs[i].r); i--)
         ;
      *r2 = obs[i].r;
      rval = 0;
      }
   return( rval);
}

#define LOG_10 2.302585

double calc_obs_magnitude( int is_comet, double obj_sun,
                                 double obj_earth, double earth_sun)
{
   double rval;

   if( is_comet)
      rval = 10. * log( obj_sun);
   else
      {
      double phase_ang;
      double phi1, phi2, log_tan_half_phase;
      const double slope_param = .15;

      phase_ang = obj_sun * obj_sun +
                     obj_earth * obj_earth - earth_sun * earth_sun;
      phase_ang /= 2. * obj_earth * obj_sun;
      if( phase_ang > .999) phase_ang = .999;
      if( phase_ang <-.999) phase_ang = -.999;
      phase_ang = acos( phase_ang);
      log_tan_half_phase = log( sin( phase_ang / 2.) / cos( phase_ang / 2.));
      phi1 = exp( -3.33 * exp( log_tan_half_phase * 0.63));
      phi2 = exp( -1.87 * exp( log_tan_half_phase * 1.22));
      rval = 5. * log( obj_sun) - 2.5 *
                  log( (1. - slope_param) * phi1 + slope_param * phi2);
      }
   rval += 5. * log( obj_earth);
   rval /= LOG_10;         /* allow for common logs,  not naturals */
   return( rval);
}

double calc_absolute_magnitude( const OBSERVE FAR *obs, int n_obs)
{
   int i, n_mags = 0;
   double rval = 0.;

   while( n_obs--)
      {
      if( obs->obs_mag)
         {
         double earth_sun = 0.;

         for( i = 0; i < 3; i++)
            earth_sun += obs->obs_posn[i] * obs->obs_posn[i];
         earth_sun = sqrt( earth_sun);
         if( earth_sun && obs->r)
            {
            rval += obs->obs_mag - calc_obs_magnitude( obs->is_comet,
                                 obs->solar_r, obs->r, earth_sun);
            n_mags++;
            }
         }
      obs++;
      }
   if( n_mags)
      rval /= (double)n_mags;
   return( rval);
}


static const char *month_names[12] = { "Jan", "Feb", "Mar", "Apr", "May",
            "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

/* observation_summary_data( ) produces the final line in an MPC report,
   such as 'From 20 observations 1997 Oct. 20-22;  RMS error .345 arcseconds'.
*/

void observation_summary_data( char *obuff, const OBSERVE FAR *obs,
                                      const int n_obs)
{
   int month, month2, day1, day2, i, n_included;
   long year, year2;

   for( i = 0; i < n_obs && !obs[i].is_included; i++)
      ;
   day1 = (int)decimal_day_to_dmy( obs[i].jd, &year,  &month );
   for( i = n_obs - 1; i >= 0 && !obs[i].is_included; i--)
      ;
   day2 = (int)decimal_day_to_dmy( obs[i].jd, &year2, &month2);

   for( i = n_included = 0; i < n_obs; i++)
      n_included += obs[i].is_included;
   sprintf( obuff, "From %d observations ", n_included);
   obuff += strlen( obuff);
   if( year == year2 && month == month2)
      sprintf( obuff, "%ld %s. %d-%d;", year, month_names[month - 1], day1, day2);
   else if( year == year2)
      sprintf( obuff, "%ld %s. %d-%s. %d;", year, month_names[month - 1],
                             day1, month_names[month2 - 1], day2);
   else
      sprintf( obuff, "%ld %s. %d-%ld %s. %d;", year, month_names[month - 1],
                             day1, year2, month_names[month2 - 1], day2);
   obuff += strlen( obuff);

   sprintf( obuff, "   RMS error %.3lf arcseconds",
             compute_rms( obs, n_obs, 1));
}
