#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifndef __GNUC__
#include <conio.h>
#include <io.h>
#else
#include <unistd.h>
#endif
#include "watdefs.h"
#include "lunar.h"
#include "afuncs.h"

#define PI 3.14159265358979323
#define N_PERTURB 14
#define J2000 2451545.
#define GAUSS_K .01720209895
#define ACCEL0 (GAUSS_K * GAUSS_K)
#define AU_PER_JRAD ((44382. * 1.609) / 149.6e+6)

#define CACHE_NOT_IN_USE 0
#define CACHE_GROWING 1
#define CACHE_USING 2

extern int perturbers;
int cache_state = 0;
FILE *cache_file;

void equatorial_to_ecliptic( double FAR *vect);       /* mpc_obs.cpp */
int planet_posn( const int planet_no, const double jd, double *vect_2000);
                                                /* mpc_obs.cpp */
int earth_lunar_posn( const double jd, double FAR *earth_loc, double FAR *lunar_loc);


int reset_cache( int action, int size)
{
   static const char *cache_file_name = "cache.dat";

   switch( action)
      {
      case CACHE_NOT_IN_USE:
         if( cache_file)
            {
            fclose( cache_file);
            cache_file = NULL;
            }
         unlink( cache_file_name);
         break;
      case CACHE_GROWING:
         cache_file = fopen( cache_file_name, "w+b");
         break;
      case CACHE_USING:
         break;
      }
   cache_state = action;
   if( cache_file)
      fseek( cache_file, 0L, SEEK_SET);
   return( 0);
}

      /* The following value for Earth's J2 comes from Danby, p 432... */
      /* but it's in units of the Earth's radius squared,  and must */
      /* be converted to AU squared. */

#define J2_IN_EARTH_UNITS (1.082e-3)
#define AU_IN_KM 1.4959787e+8
#define EARTH_R 6378.140
#define EARTH_J2 (J2_IN_EARTH_UNITS*EARTH_R * EARTH_R / (AU_IN_KM * AU_IN_KM))

void analytical_j2_gradient( double *grad, double x, double y, double z)
{
   double r2 = x * x + y * y + z * z;
   double r7 = r2 * r2 * r2 * sqrt( r2);
   double tval = (3. * r2 - 15. * z * z) / r7;

   grad[0] = x * tval;
   grad[1] = y * tval;
   grad[2] = z * (6. * r2 / r7 + tval);
}

      /* 22 Oct 2001:  replaced Uranus,  Neptune masses w/DE-405 values */

double planet_mass[N_PERTURB + 1] = { 1.,
         1.66e-7, 2.448e-6,            /* mercury, venus */
         2.98994e-6, 3.23e-7,          /* earth, mars */
         1.8986e+27 / 1.9891e+30,      /* Jupiter */
//       9.5479193e-4,                 /* Jupiter */
         2.85886e-4,                    /*  saturn */
//       9.549533e-4, 2.85886e-4,       /* jupiter,  saturn */
//       4.3554e-5, 5.1776e-5, 8.0e-9, /* uranus, neptune, pluto */
//       4.365792e-5, 5.1776e-5, 8.0e-9, /* uranus, neptune, pluto */
         4.366244e-5, 5.15139e-5, 8.0e-9, /* uranus, neptune, pluto */
         2.98994e-6 / 81.3,            /* moon */
         893.3E+20 / 1.9891E+30,      /* mass of Io (in kg, divided by sun) */
         479.7E+20 / 1.9891E+30,      /* similar for Europa, */
         1482E+20  / 1.9891E+30,      /* Ganymede, */
         1076E+20  / 1.9891E+30 };    /* and Callisto */

      /* If we're closer than .1 AU,  include Galileans separately: */

#define GALILEAN_LIMIT 0
//        #define GALILEAN_LIMIT .1

int calc_derivatives( double jd, double *ival, double *oval)
{
   double r = 0.;
   int i, j, local_perturbers = perturbers;
   double lunar_loc[3], jupiter_loc[3];

   oval[0] = ival[3];
   oval[1] = ival[4];
   oval[2] = ival[5];
   for( i = 0; i < 3; i++)
      r += ival[i] * ival[i];
   r *= sqrt( r);
   for( i = 0; i < 3; i++)
      oval[i + 3] = -ACCEL0 * ival[i] / r;
   if( perturbers)
      for( i = 1; i < 15; i++)
         if( i != 9 && ((local_perturbers >> i) & 1))
            {
            double planet_loc[15], accel[3], mass_to_use;

            r = 0.;
            if( cache_state == CACHE_USING)
               {
               fread( planet_loc + 12, 3, sizeof( double), cache_file);
               for( j = 12; j < 15; j++)
                  r += planet_loc[j] * planet_loc[j];
               planet_loc[2] = sqrt( r);
               r = 0.;
               }
            else
               {
               if( i > 10)       /* Galileans */
                  {
                  double matrix[10], jsat_loc[15], r2 = 0.;
//                FILE *debug_file = fopen( "debug.dat", "ab");

                  calc_jsat_loc( jd, jsat_loc, 1 << (i - 11), 0L);
                  setup_precession( matrix, 2000. + (jd - J2000)/365.25, 2000.);
                  precess_vector( matrix, jsat_loc + (i - 11) * 3,
                                                           planet_loc + 12);
                  for( j = 0; j < 3; j++)
                     {
                     planet_loc[12 + j] *= AU_PER_JRAD;
                     planet_loc[12 + j] += jupiter_loc[j];
                     r2 += planet_loc[12 + j] * planet_loc[12 + j];
                     }
                  planet_loc[2] = sqrt( r2);
//                fprintf( debug_file, "JD %.5lf, satellite %d: r = %lf\n",
//                         jd, i - 10, planet_loc[2]);
//                fclose( debug_file);
                  }
               else
                  {
                  double r2 = 0.;

                  if( local_perturbers & 1024)   /* if the moon is included */
                     {
                     if( i == 3)
                        earth_lunar_posn( jd, planet_loc, lunar_loc);
                     else if( i == 10)
                        memcpy( planet_loc, lunar_loc, 3 * sizeof( double));
                     else
                        planet_posn( i, jd, planet_loc);
                     }
                  else
                     planet_posn( i, jd, planet_loc);

                  for( j = 0; j < 3; j++)
                     r2 += planet_loc[j] * planet_loc[j];
                  memcpy( planet_loc + 12, planet_loc, 3 * sizeof( double));
                  planet_loc[2] = sqrt( r2);
                  }
               if( cache_state == CACHE_GROWING)
                  fwrite( planet_loc + 12, 3, sizeof( double), cache_file);
               }

            for( j = 0; j < 3; j++)
               {
               accel[j] = ival[j] - planet_loc[12 + j];
               r += accel[j] * accel[j];
               }

            if( i == 5 && r < GALILEAN_LIMIT * GALILEAN_LIMIT)
               {
               memcpy( jupiter_loc, planet_loc + 12, 3 * sizeof( double));
               local_perturbers |= (15 << 11);
               }

            if( i == 3 && r < .001 * .001)
               {          /* Within .001 AU,  we take J2 into account: */
               double grad[3], delta_j2000[3];
               const double obliq_2000 = 23.4392911 * PI / 180.;
               const double eff_j2 = EARTH_J2 * ACCEL0 * planet_mass[3] / 2.;

                           /* Remembering the 'accels' are 'deltas' now... */
               memcpy( delta_j2000, accel, 3 * sizeof( double));
                           /* Cvt ecliptic to equatorial: */
               rotate_vector( delta_j2000, obliq_2000, 0);
               analytical_j2_gradient( grad, delta_j2000[0],
                                       delta_j2000[1], delta_j2000[2]);
                           /* Cvt equatorial to ecliptic: */
               rotate_vector( grad, -obliq_2000, 0);
                           /* And add 'em to the output acceleration: */
               for( j = 0; j < 3; j++)
                  oval[j + 3] += eff_j2 * grad[j];
               }

            mass_to_use = planet_mass[i];
                     /* If we're including the earth,  but the moon isn't */
                     /* being included separately,  add the mass of the moon:*/
            if( i == 3 && !(local_perturbers & 1024))
               mass_to_use += planet_mass[10];
                     /* Similarly for Jupiter and the Galileans: */
            if( i == 5 && !(local_perturbers & 2048))
               mass_to_use += planet_mass[11] + planet_mass[12]
                            + planet_mass[13] + planet_mass[14];
            r = -ACCEL0 * mass_to_use / (r * sqrt( r));
            for( j = 0; j < 3; j++)
               oval[j + 3] += r * accel[j];
            r = planet_loc[2];
            r = -ACCEL0 * mass_to_use / (r * r * r);
            for( j = 0; j < 3; j++)
               oval[j + 3] += r * planet_loc[j + 12];
            }
   return( 0);
}

int take_step( double jd, double *ival, double *ovals, double *errs,
                               int n_vals, double step)
{
   double *ivals[7], *ivals_p[6];
   int i, j, k;
   double bvals[27] = {2. / 9.,
            1. / 12., 1. / 4.,
            69. / 128., -243. / 128., 135. / 64.,
            -17. / 12., 27. / 4., -27. / 5., 16. / 15.,
            65. / 432., -5. / 16., 13 / 16., 4 / 27., 5. / 144.,
            47. / 450., 0., 12 / 25., 32. / 225., 1. / 30., 6. / 25.,
            -1. / 150., 0., .03, -16. / 75., -.05, .24};
   double *bptr = bvals;
   double avals[7] = { 0., 2. / 9., 1./3., .75, 1., 5./6. };

   ivals[1] = (double *)calloc( 12 * n_vals, sizeof( double));
   for( i = 0; i < 6; i++)
      {
      ivals[i + 1] = ivals[1] + i * n_vals;
      ivals_p[i] = ivals[1] + (i + 6) * n_vals;
      }

   calc_derivatives( jd, ival, ivals_p[0]);

   for( j = 1; j < 7; j++)
      {
      for( i = 0; i < n_vals; i++)
         {
         double tval = 0.;

         for( k = 0; k < j; k++)
            tval += bptr[k] * ivals_p[k][i];
         ivals[j][i] = tval * step + ival[i];
         }
      bptr += j;
      if( j != 6)
         calc_derivatives( jd + step * avals[j], ivals[j], ivals_p[j]);
      }

   if( errs)
      for( i = 0; i < n_vals; i++)
         {
         double tval = 0.;

         for( k = 0; k < 6; k++)
            tval += bptr[k] * ivals_p[k][i];
         errs[i] = step * tval;
         }

   memcpy( ovals, ivals[6], n_vals * sizeof( double));
   free( ivals[1]);
   return( 0);
}

int full_rk_step( double *ivals, double *ovals, double t0, double t1,
                              int n_vals, double max_err)
{
#ifdef CANT_USE_THIS
   double step = t1 - t0;
   double *errs = (double *)calloc( 2 * n_vals, sizeof( double));
   double *new_vals = errs + n_vals;

   if( !errs)
      return( -1);
   memcpy( ovals, ivals, n_vals * sizeof( double));
   max_err *= max_err;
   while( t0 != t1)
      {
      double err_val = 0.;
      const double chicken_factor = .9;
      int i;

      take_step( t0, ovals, new_vals, errs, n_vals, step);
      for( i = 0; i < n_vals; i++)
         err_val += errs[i] * errs[i];
      if( err_val < max_err)   /* yeah,  it was a good step */
         {
         memcpy( ovals, new_vals, n_vals * sizeof( double));
         t0 += step;
         }
      step *= chicken_factor * exp( log( max_err / err_val) / 5.);
      if( t0 < t1)
         if( t0 + step > t1)
            step = t1 - t0;
      if( t1 < t0)
         if( t0 + step < t1)
            step = t1 - t0;
      }
   free( errs);
#else
   take_step( t0, ivals, ovals, NULL, n_vals, t1 - t0);
#endif
   return( 0);
}

int find_best_fit_planet( const double jd, const double *ivect,
                                 double *rel_vect, const int included)
{
   int i, j, rval = 0;
   double best_fit = 0., vel[3];

// memcpy( rel_vect, ivect, 6 * sizeof( double));
// return( 0);

   for( i = 0; i < 9; i++)
      if( (included >> i) & 1)
         {
         double planet_loc[3], delta[3], dist2 = 0., curr_fit;

         if( i == 3)
            earth_lunar_posn( jd, planet_loc, NULL);
         else
            planet_posn( i, jd, planet_loc);
         for( j = 0; j < 3; j++)
            {
            delta[j] = ivect[j] - planet_loc[j];
            dist2 += delta[j] * delta[j];
            }

         curr_fit = planet_mass[i] / exp( log( dist2) * 1.25);
         if( !i || curr_fit > best_fit)
            {
            rval = i;
            best_fit = curr_fit;
            memcpy( rel_vect, delta, 3 * sizeof( double));
            }
         }

   if( !rval)        /* is heliocentric after all */
      memcpy( rel_vect + 3, ivect + 3, 3 * sizeof( double));
   else
      {
#ifdef DEBUGGING_RUNGE
      FILE *debug_file = fopen( "debug.dat", "ab");
#endif

      for( i = 0; i < 2; i++)
         {
         double planet_loc[3];
         double delta = 1. / 1440.;    /* one minute delta... */

         if( i == 3)
            earth_lunar_posn( jd + (i ? delta : -delta), planet_loc, NULL);
         else
            planet_posn( rval, jd + (i ? delta : -delta), planet_loc);

         if( i)
            for( j = 0; j < 3; j++)
               {
               vel[j] = (planet_loc[j] - vel[j]) / (2. * delta);
               rel_vect[j + 3] = ivect[j + 3] - vel[j];
               }
         else
            memcpy( vel, planet_loc, 3 * sizeof( double));
         }
      }
   return( rval);
}

#ifdef TEST_CODE
void main( int argc, char **argv)
{
   double x[6] = {1., .1, -.1, -.1, 1., .2};
   double step = atof( argv[1]), t = 0., t_end = 0.;
   double tl = 1.e-7;
   int step_no = 0, step_freq = 1;
   int i, quit = 0;

   if( argc > 2)
      t_end = atof( argv[2]);
   for( i = 2; i < argc; i++)
      switch( argv[i][0])
         {
         case 's':
            step_freq = atoi( argv[i] + 1);
            break;
         case 'l':
            tl = atof( argv[i] + 1);
            break;
         default:
            break;
         }

   while( !quit)
      {
      double max_te = 0., te_arr[6];
      int i;

      if( argc > 3 && step < 0.)
         exit( 0);
      if( t_end)
         if( t_end - t < step)
            {
            step = t_end - t;
            quit = 1;
            }
      take_step( x, x, te_arr, 6, step);
      t += step;
      step_no++;
      for( i = 0; i < 6; i++)
         if( fabs( te_arr[i]) > max_te)
            max_te = fabs( te_arr[i]);
      step *= .9 * exp( log( tl / max_te) / 5.);
      if( kbhit( ))
         if( getch( ) == 27)
            quit = 1;
      if( quit || !(step_no % step_freq))
         printf( "%15.12lf %15.12lf %15.12lf  %10lf %10lf\n",
                                    x[0], x[1], x[2], t, step);
      }
}
#endif
